# -*- coding: utf-8 -*-
# BioSTEAM: The Biorefinery Simulation and Techno-Economic Analysis Modules
# Copyright (C) 2020, Yoel Cortes-Pena <yoelcortes@gmail.com>
# 
# This module is under the UIUC open-source license. See 
# github.com/BioSTEAMDevelopmentGroup/biosteam/blob/master/LICENSE.txt
# for license details.
"""
"""
from . import mixture
from . import ideal_mixture_model
from . import mixture_builders

__all__ = (*mixture.__all__,
           *ideal_mixture_model.__all__,
           *mixture_builders.__all__,           
)

from .mixture import *
from .ideal_mixture_model import *
from .mixture_builders import *



# -*- coding: utf-8 -*-
# BioSTEAM: The Biorefinery Simulation and Techno-Economic Analysis Modules
# Copyright (C) 2020, Yoel Cortes-Pena <yoelcortes@gmail.com>
# 
# A significant portion of this module originates from:
# Chemical Engineering Design Library (ChEDL). Utilities for process modeling.
# Copyright (C) 2020 Caleb Bell <Caleb.Andrew.Bell@gmail.com>
# 
# This module is under a dual license:
# 1. The UIUC open-source license. See 
# github.com/BioSTEAMDevelopmentGroup/biosteam/blob/master/LICENSE.txt
# for license details.
# 
# 2. The MIT open-source license. See
# https://github.com/CalebBell/thermo/blob/master/LICENSE.txt for details.
from ..functional import horner
from ..base import (InterpolatedTDependentModel, 
                    TPDependentHandleBuilder, 
                    TDependentModel, 
                    PhaseTPHandleBuilder,
                    functor)
from .data import (VDI_saturation_dict,
                   lookup_VDI_tabular_data,
                   mu_data_Dutt_Prasad,
                   mu_data_VN3,
                   mu_data_VN2,
                   mu_data_Perrys_8E_2_313,
                   mu_data_Perrys_8E_2_312,
                   mu_data_VDI_PPDS_7,
                   mu_data_VDI_PPDS_8,
)
from .thermal_conductivity import IAPWS_rho_hook
from chemicals import viscosity
from chemicals.dippr import EQ101, EQ102


mul = 'mu.l'
mu_IAPWS = functor(viscosity.mu_IAPWS, mul)
mu_IAPWS.functor.hook = IAPWS_rho_hook
Viswanath_Natarajan_2 = functor(viscosity.Viswanath_Natarajan_2, mul)
Viswanath_Natarajan_3 = functor(viscosity.Viswanath_Natarajan_3, mul)
Letsou_Stiel = functor(viscosity.Letsou_Stiel, mul)
Przedziecki_Sridhar = functor(viscosity.Przedziecki_Sridhar, mul)
PPDS9 = functor(viscosity.PPDS9, mul)
Lucas = functor(viscosity.Lucas, mul)

def Przedziecki_Sridhar_hook(self, T, kwargs):
    Vm = kwargs['Vm']
    if callable(Vm):
        kwargs = kwargs.copy()
        kwargs['Vm'] = Vm.at_T(T)
    return self.function(T, **kwargs)
Przedziecki_Sridhar.functor.hook = Przedziecki_Sridhar_hook

def Lucas_hook(self, T, P, kwargs):
    kwargs = kwargs.copy()
    Psat = kwargs['Psat']
    mu_l = kwargs['mu_l']
    if callable(Psat):
        kwargs['Psat'] = Psat(T)
    if callable(mu_l):
        kwargs['mu_l'] = mu_l.at_T(T)
    return self.function(T, P, **kwargs)
Lucas.functor.hook = Lucas_hook

@TPDependentHandleBuilder('mu.l')
def viscosity_liquid_handle(handle, CAS, MW, Tm, Tc, Pc, Vc, omega, Psat, Vl):
    add_model = handle.add_model
    if CAS in VDI_saturation_dict:
        Ts, Ys = lookup_VDI_tabular_data(CAS, 'Mu (l)')
        model = InterpolatedTDependentModel(Ts, Ys, Ts[0], Ts[-1], 
                                            name='VDI-interpolated')
        add_model(model)
    if CAS in mu_data_Dutt_Prasad:
        A, B, C, Tmin, Tmax = mu_data_Dutt_Prasad[CAS]
        data = (A - 3.0, B, C)
        add_model(Viswanath_Natarajan_3.functor.from_args(data), Tmin, Tmax)
    if CAS in mu_data_VN3:
        A, B, C, Tmin, Tmax = mu_data_VN3[CAS]
        data = (A - 3.0, B, C)
        add_model(Viswanath_Natarajan_3.functor.from_args(data), Tmin, Tmax)
    if CAS in mu_data_VN2:
        A, B, Tmin, Tmax = mu_data_VN2[CAS]
        data = (A, B)
        add_model(Viswanath_Natarajan_2.functor.from_args(data), Tmin ,Tmax)
    if CAS in mu_data_Perrys_8E_2_313:
        C1, C2, C3, C4, C5, Tmin, Tmax = mu_data_Perrys_8E_2_313[CAS]
        data = (C1, C2, C3, C4, C5)
        add_model(EQ101.functor.from_args(data), Tmin, Tmax)
    if CAS in mu_data_VDI_PPDS_7:
        add_model(PPDS9.functor.from_args(mu_data_VDI_PPDS_7[CAS]))
    data = (MW, Tc, Pc, omega)
    if all(data):
        add_model(Letsou_Stiel.functor.from_args(data), Tc/4, Tc)
    # data = (Tm, Tc, Pc, Vc, Vl, omega, MW)
    # if all(data):
    #     # Pretty bad model; so do not use
    #     add_model(Przedziecki_Sridhar.functor.from_args(data), Tm, Tc)
    data = (Tc, Pc, omega)
    if all(data):
        data = (Tc, Pc, omega, Psat, handle)
        add_model(Lucas.functor.from_args(data), Tm, Tc,
                  name='Lucas')
viscosity.viscosity_liquid_handle = viscosity_liquid_handle

mug = 'mu.g'
Yoon_Thodos = functor(viscosity.Yoon_Thodos, mug)
Stiel_Thodos = functor(viscosity.Stiel_Thodos, mug)
Lucas_gas = functor(viscosity.Lucas_gas, mug)
viscosity_gas_Gharagheizi = functor(viscosity.viscosity_gas_Gharagheizi, mug)

@TPDependentHandleBuilder('mu.g')
def viscosity_gas_handle(handle, CAS, MW, Tc, Pc, Zc, dipole):
    add_model = handle.add_model
    if CAS in VDI_saturation_dict:
        Ts, Ys = lookup_VDI_tabular_data(CAS, 'Mu (g)')
        Tmin = Ts[0]
        Tmax = Ts[-1]
        model = InterpolatedTDependentModel(Ts, Ys, Tmin, Tmax,
                                            name='VDI-interpolated')
        add_model(model)
    if CAS in mu_data_Perrys_8E_2_312:
        C1, C2, C3, C4, Tmin, Tmax = mu_data_Perrys_8E_2_312[CAS]
        data = (C1, C2, C3, C4)
        add_model(EQ102.functor.from_args(data), Tmin, Tmax)
    if CAS in mu_data_VDI_PPDS_8:
        data = mu_data_VDI_PPDS_8[CAS].tolist()
        data.reverse()
        add_model(horner.functor.from_kwargs({'coeffs':data}),
                  name='VDI-PPDS')
    # data = (Tc, Pc, Zc, MW)
    # if all(data):
    #     Tmin = 0; Tmax = 1e3
    #     add_model(Lucas_gas.from_args(data), Tmin, Tmax,
    #               name='Lucas')
    data = (Tc, Pc, MW)
    if all(data):
        Tmin = 0; Tmax = 5e3
        add_model(viscosity_gas_Gharagheizi.functor.from_args(data), Tmin, Tmax,
                  name='Gharagheizi')
        add_model(Yoon_Thodos.functor.from_args(data), Tmin, Tmax)
        add_model(Stiel_Thodos.functor.from_args(data), Tmin, Tmax)
        # Intelligently set limit
        # GHARAGHEIZI turns nonsensical at ~15 K, YOON_THODOS fine to 0 K,
        # same as STIEL_THODOS
viscosity.viscosity_gas_handle = viscosity_gas_handle

viscosity.viscosity_handle = PhaseTPHandleBuilder('mu', None,
                                                  viscosity_liquid_handle,
                                                  viscosity_gas_handle)



# -*- coding: utf-8 -*-
# BioSTEAM: The Biorefinery Simulation and Techno-Economic Analysis Modules
# Copyright (C) 2020, Yoel Cortes-Pena <yoelcortes@gmail.com>
# 
# This module is under the UIUC open-source license. See 
# github.com/BioSTEAMDevelopmentGroup/biosteam/blob/master/LICENSE.txt
# for license details.
"""
"""
import flexsolve as flx
from .. import functional as fn
from ..exceptions import InfeasibleRegion
from .fugacity_coefficients import IdealFugacityCoefficients
import numpy as np

__all__ = ('solve_x', 'solve_y')

def x_iter(x, x_gamma_poyinting, gamma, poyinting, T):
    x = fn.normalize(x)
    # Add back trace amounts for activity coefficients at infinite dilution
    mask = x < 1e-32
    x[mask] = 1e-32
    denominator = gamma(x, T) * poyinting(x, T)
    try:
        x = x_gamma_poyinting / denominator
    except FloatingPointError: 
        raise InfeasibleRegion('liquid phase composition')
    if (np.abs(x) > 1e16).any():
        raise InfeasibleRegion('liquid phase composition')
    return x

def solve_x(x_gamma_poyinting, gamma, poyinting, T, x_guess):
    if x_guess is None: x_guess = x_gamma_poyinting
    args = (x_gamma_poyinting, gamma, poyinting, T)
    try:
        x = flx.aitken(x_iter, x_guess, 1e-10, args=args, checkiter=False)
    except InfeasibleRegion:
        x = x_gamma_poyinting
    return x
        
def y_iter(y, y_phi, phi, T, P):
    y = fn.normalize(y)
    return y_phi / phi(y, T, P)

def solve_y(y_phi, phi, T, P, y_guess):
    if isinstance(phi, IdealFugacityCoefficients): return y_phi
    elif y_guess is None: y_guess = y_phi
    return flx.aitken(y_iter, y_phi, 1e-9, args=(y_phi, phi, T, P),
                      checkiter=False)


# -*- coding: utf-8 -*-
# BioSTEAM: The Biorefinery Simulation and Techno-Economic Analysis Modules
# Copyright (C) 2020, Yoel Cortes-Pena <yoelcortes@gmail.com>
# 
# This module is under the UIUC open-source license. See 
# github.com/BioSTEAMDevelopmentGroup/biosteam/blob/master/LICENSE.txt
# for license details.
"""
"""
from ..registry import Registry

__all__ = ('registered',)


def registered(ticket_name):
    return lambda cls: _registered(cls, ticket_name)

def _registered(cls, ticket_name):
    cls.registry = Registry()
    cls.ticket_name = ticket_name
    cls.ticket_number = 0
    cls.unregistered_ticket_number = 0
    cls._take_unregistered_ticket = _take_unregistered_ticket
    cls._take_ticket = _take_ticket
    cls._unregister = _unregister
    cls._register = _register
    cls.ID = ID
    cls.__repr__ = __repr__
    cls.__str__ = __str__
    return cls

@classmethod
def _take_ticket(cls):
    cls.ticket_number += 1
    return cls.ticket_name + str(cls.ticket_number)
@classmethod
def _take_unregistered_ticket(cls):
    if cls.unregistered_ticket_number > 100:
       cls.unregistered_ticket_number = 1 
    else:
        cls.unregistered_ticket_number += 1
    return cls.ticket_name + '.' + str(cls.unregistered_ticket_number)

def _unregister(self):
    try: delattr(self.registry, self._ID)
    except: pass
    
def _register(self, ID):
    if ID:
        self.registry[ID] = self
    elif ID == "":
        ID = self._take_ticket()
        self.registry.__dict__[ID] = self
    else:
        ID = self._take_unregistered_ticket()
    self._ID = ID
        
@property
def ID(self):
    """Unique identification (str). If set as '', it will choose a default ID."""
    return self._ID

@ID.setter
def ID(self, ID):
    self._unregister()
    self._register(ID)
    
def __repr__(self):
    return f"<{type(self).__name__}: {self.ID}>"

def __str__(self):
    return self.ID



# -*- coding: utf-8 -*-
# BioSTEAM: The Biorefinery Simulation and Techno-Economic Analysis Modules
# Copyright (C) 2020, Yoel Cortes-Pena <yoelcortes@gmail.com>
# 
# This module is under the UIUC open-source license. See 
# github.com/BioSTEAMDevelopmentGroup/biosteam/blob/master/LICENSE.txt
# for license details.
"""
"""
from .utils import colors
import flexsolve as flx

__all__ = ('UndefinedChemical',
           'UndefinedPhase',
           'DimensionError',
           'InfeasibleRegion',
           'DomainError',
           'InvalidMethod',
           'message_with_object_stamp',
           'try_method_with_object_stamp',
           'raise_error_with_object_stamp')

class InfeasibleRegion(RuntimeError):
    """Runtime error regarding infeasible processes."""
    def __init__(self, region): 
        self.region = region
        super().__init__(region + ' is infeasible')

class UndefinedChemical(Exception):
    """Exception regarding undefined chemicals."""
    def __init__(self, ID): super().__init__(repr(ID))
    
class UndefinedPhase(Exception):
    """Exception regarding undefined phases."""
    def __init__(self, phase): super().__init__(repr(phase))

class DimensionError(ValueError):
    """ValueError regarding wrong dimensions."""

class DomainError(ValueError):
    """ValueError regarding an attempt to evaluate a model out of its domain."""
    def __init__(self, msg, **data):
        super().__init__(msg)
        self.__dict__.update(data)

class InvalidMethod(ValueError):
    """ValueError regarding an attempt to evaluate an invalid method."""
    def __init__(self, method):
        super().__init__(repr(method))
    
def message_with_object_stamp(object, msg):
    object_name = str(repr(object))
    if object_name in msg:
        return msg
    else:
        return colors.violet(object_name) + ' ' + msg

def raise_error_with_object_stamp(object, error):
    try: 
        msg, *args = error.args
        error.args = (message_with_object_stamp(object, msg), *args)
    except: pass
    raise error

def try_method_with_object_stamp(object, method, args=()):
    try:
        return method(*args)
    except KeyError as error:
        raise error
    except Exception as error:
        raise_error_with_object_stamp(object, error)


# -*- coding: utf-8 -*-
# BioSTEAM: The Biorefinery Simulation and Techno-Economic Analysis Modules
# Copyright (C) 2020, Yoel Cortes-Pena <yoelcortes@gmail.com>
# 
# This module is under the UIUC open-source license. See 
# github.com/BioSTEAMDevelopmentGroup/biosteam/blob/master/LICENSE.txt
# for license details.
"""
"""
import flexsolve as flx
from ..utils import Cache
from .equilibrium import Equilibrium
from .activity_coefficients import IdealActivityCoefficients
from chemicals import solubility_eutectic
import numpy as np

__all__ = ('SLE', 'SLECache')

class SLE(Equilibrium, phases='ls'):
    """
    Create an SLE object that performs solid-liquid equilibrium for a given solute
    when called with a temperature and pressure.
        
    Parameters
    ----------
    imol=None : :class:`~thermosteam.indexer.MaterialIndexer`, optional
        Molar chemical phase data is stored here.
    thermal_condition=None : :class:`~thermosteam.ThermalCondition`, optional
        The temperature and pressure used in calculations are stored here.
    thermo=None : :class:`~thermosteam.Thermo`, optional
        Themodynamic property package for equilibrium calculations.
        Defaults to `thermosteam.settings.get_thermo()`.
    activity_coefficient=None : float
        Activity coefficient of solute in the liquid; only valid if
        `thermo.Gamma` is :class:`~thermosteam.equilibrium.activity_coefficients.IdealActivityCoefficients`.
    
    Examples
    --------
    Solve SLE of glucose in water:
    
    >>> from thermosteam import Chemical, indexer, equilibrium, settings
    >>> Glucose = Chemical('Glucose', Tm=419.15, Hfus=19930)
    >>> Glucose.Cn.s.add_model(224.114064, top_priority=True)
    >>> Glucose.Cn.l.add_model(360.312, top_priority=True) # More or less in solution
    >>> settings.set_thermo(['Water', Glucose], cache=True)
    >>> imol = indexer.MolarFlowIndexer(l=[('Water', 10), ('Glucose', 1)], phases=('s', 'l'))
    >>> sle = equilibrium.SLE(imol)
    >>> sle('Glucose', T=298.15) 
    >>> sle
    SLE(imol=MolarFlowIndexer(
            l=[('Water', 10), ('Glucose', 0.01308)],
            s=[('Glucose', 0.9869)]),
        thermal_condition=ThermalCondition(T=298.15, P=101325))
        
    Results may not be too accurate sometimes, but the solubility 
    (mol fraction of solute in solvent) may be specified:
    
    >>> sle('Glucose', T=298.15, solubility=0.0833)
    >>> sle
    SLE(imol=MolarFlowIndexer(
            l=[('Water', 10), ('Glucose', 0.9087)],
            s=[('Glucose', 0.09131)]),
        thermal_condition=ThermalCondition(T=298.15, P=101325))
    
    Solve SLE of tetradecanol in octanol:
        
    >>> from thermosteam import indexer, equilibrium, settings
    >>> settings.set_thermo(['Methanol', 'Tetradecanol'], cache=True)
    >>> imol = indexer.MolarFlowIndexer(l=[('Methanol', 10), ('Tetradecanol', 30)], phases=('s', 'l'))
    >>> sle = equilibrium.SLE(imol)
    >>> sle('Tetradecanol', T=300)
    >>> sle
    SLE(imol=MolarFlowIndexer(
            l=[('Methanol', 10), ('Tetradecanol', 19.06)],
            s=[('Tetradecanol', 10.94)]),
        thermal_condition=ThermalCondition(T=300.00, P=101325))
    
    Solve SLE of pure tetradecanol:
    
    >>> from thermosteam import indexer, equilibrium, settings
    >>> settings.set_thermo(['Octanol', 'Tetradecanol'], cache=True)
    >>> imol = indexer.MolarFlowIndexer(l=[('Tetradecanol', 30)], phases=('s', 'l'))
    >>> sle = equilibrium.SLE(imol)
    >>> sle('Tetradecanol', T=300) # Under melting point
    >>> sle
    SLE(imol=MolarFlowIndexer(phases=('l', 's'),
            s=[('Tetradecanol', 30)]),
        thermal_condition=ThermalCondition(T=300.00, P=101325))
    >>> sle('Tetradecanol', T=320) # Over melting point
    >>> sle
    SLE(imol=MolarFlowIndexer(phases=('l', 's'),
            l=[('Tetradecanol', 30)]),
        thermal_condition=ThermalCondition(T=320.00, P=101325))      
    
    
    """
    __slots__ = ('_x', # [float] Fraction of solute as a solid.
                 '_gamma', # [ActivityCoefficients] Estimates activity coefficients of a liquid.
                 '_liquid_mol', # [1d array] Liquid molar data.
                 '_solid_mol', # [1d array] Solid molar data.
                 '_phase_data', # tuple[str, 1d array] Phase-data pairs.
                 '_index', # [1d array] Index of chemicals in equilibrium.
                 '_chemical', # [Chemical] Single chemical in equilibrium.
                 '_nonzero', # [1d array(bool)] Chemicals present in the mixture
                 '_mol_solute', # [float] Solute molar data.
                 '_solute_index', # [int] Solute index
                 '_solute_gamma_index', # [int] Solute index for activity coefficients
                 'activity_coefficient', # [float] Activity coefficient of solute in the liquid; only valid thermo.Gamma is IdealActivityCoefficients.
    )
    
    def __init__(self, imol=None, thermal_condition=None, thermo=None, 
                 solubility_weight=None, activity_coefficient=None):
        super().__init__(imol, thermal_condition, thermo)
        imol = self._imol
        self._phase_data = tuple(imol)
        self._liquid_mol = liquid_mol = imol['l']
        self._solid_mol = imol['s']
        self._nonzero = np.zeros(liquid_mol.shape, dtype=bool)
        self._index = ()
        self._chemical = None
        self._x = None
        self.activity_coefficient = activity_coefficient
    
    def _setup(self):
        # Get flow rates
        liquid_mol = self._liquid_mol
        solid_mol = self._solid_mol
        mol = liquid_mol + solid_mol
        solute_index = self._solute_index
        self._mol_solute = mol_solute = mol[solute_index]
        if not mol_solute:
            raise RuntimeError('no solute available')
        nonzero = mol > 0
        if (self._nonzero == nonzero).all():
            index = self._index
        else:
            chemicals = self.chemicals
            # Set up indices for both equilibrium and non-equilibrium species
            index = chemicals.get_lle_indices(nonzero)   
            N = len(index)
            if N == 1:
                self._chemical = chemicals.tuple[solute_index]
            else:
                # Set equilibrium objects
                eq_chems = chemicals.tuple
                eq_chems = [eq_chems[i] for i in index]
                self._nonzero = nonzero
                self._index = index
                thermo = self._thermo
                self._gamma = thermo.Gamma(eq_chems)
                self._solute_gamma_index = self._index.index(solute_index)
        
    def __call__(self, solute, T, P=None, solubility=None):
        """
        Perform solid-liquid equilibrium.

        Parameters
        ----------
        T : float
            Operating temperature [K].
        P : float, optional
            Operating pressure [Pa].
        solubility : float, optional
            Mol fraction of solute at maximum solubility.
            
        """
        thermal_condition = self.thermal_condition
        if P: thermal_condition.P = P
        thermal_condition.T = T
        chemicals = self.chemicals
        self._solute_index = solute_index = chemicals.get_index(solute)
        if solubility is not None:
            self._index = slice(None)
            self._update_solubility(solubility)
            return
        self._setup()
        liquid_mol = self._liquid_mol
        solid_mol = self._solid_mol
        if self._chemical:
            Tm = self._chemical.Tm
            if T > Tm:
                liquid_mol[solute_index] = self._mol_solute
                solid_mol[solute_index] = 0.
            else:
                liquid_mol[solute_index] = 0.
                solid_mol[solute_index] = self._mol_solute
        else:
            solubility = self._solve_x(T)
            self._update_solubility(solubility)
    
    def _update_solubility(self, x):
        solute_index = self._solute_index
        liquid_mol = self._liquid_mol
        solid_mol = self._solid_mol
        F_mol_liquid = liquid_mol[self._index].sum() - liquid_mol[solute_index]
        mol_solute = self._mol_solute
        x_max = mol_solute / (F_mol_liquid + mol_solute)
        if x < 0.:
            liquid_mol[solute_index] = 0.
            solid_mol[solute_index] = mol_solute
        elif x >= x_max:
            liquid_mol[solute_index] = mol_solute
            solid_mol[solute_index] = 0.
        else:
            liquid_mol[solute_index] = mol_solute_liquid = F_mol_liquid * x / (1 - x)
            solid_mol[solute_index] = mol_solute - mol_solute_liquid 
    
    def _solve_x(self, T):
        solute_chemical = self.chemicals.tuple[self._solute_index]
        Tm = solute_chemical.Tm
        if Tm is None: raise RuntimeError(f"solute {solute_chemical} does not have a melting temperature, Tm")
        Cpl = solute_chemical.Cn.l(T)
        Cps = solute_chemical.Cn.s(T)
        Hm = solute_chemical.Hfus
        if Hm is None: raise RuntimeError(f"solute {solute_chemical} does not have a heat of fusion, Hfus")
        gamma = 1.
        x = solubility_eutectic(T, Tm, Hm, Cpl, Cps, gamma) # Initial guess
        args = (T, Tm, Hm, Cpl, Cps)
        if isinstance(self._gamma, IdealActivityCoefficients):
            return solubility_eutectic(T, Tm, Hm, Cpl, Cps, self.activity_coefficient or 1.)
        return flx.aitken(self._x_iter, x, xtol=1e-6, args=args, checkiter=False, maxiter=100)
        
    def _x_iter(self, x, T, Tm, Hm, Cpl, Cps):
        self._update_solubility(x)
        liquid_mol = self._liquid_mol[self._index]
        F_mol_liquid = liquid_mol.sum()
        x_l = liquid_mol / F_mol_liquid
        gamma = self._gamma(x_l, T)
        return solubility_eutectic(T, Tm, Hm, Cpl, Cps, gamma[self._solute_gamma_index])
        
        
class SLECache(Cache): load = SLE
del Cache, Equilibrium     


# -*- coding: utf-8 -*-
# BioSTEAM: The Biorefinery Simulation and Techno-Economic Analysis Modules
# Copyright (C) 2020, Yoel Cortes-Pena <yoelcortes@gmail.com>
# 
# This module is under the UIUC open-source license. See 
# github.com/BioSTEAMDevelopmentGroup/biosteam/blob/master/LICENSE.txt
# for license details.
"""
"""
__all__ = ('PoyintingCorrectionFactors',
           'IdealPoyintingCorrectionFactors')

class PoyintingCorrectionFactors:
    """Abstract class for the estimation of Poyinting correction factors. Non-abstract subclasses should implement the following methods:
        
    __init__(self, chemicals: Iterable[:class:`~thermosteam.Chemical`]):
        Should use pure component data from chemicals to setup future calculations of Poyinting correction factors.
    
    __call__(self, y: 1d array, T: float):
        Should accept an array of vapor molar compositions `y`, and temperature `T` (in Kelvin), and return an array of Poyinting correction factors. Note that the molar compositions must be in the same order as the chemicals defined when creating the PoyintingCorrectionFactors object.
        
    """
    __slots__ = ()
    
    def __init__(self, chemicals):
        self.chemicals = chemicals
    
    def __repr__(self):
        chemicals = ", ".join([i.ID for i in self.chemicals])
        return f"<{type(self).__name__}([{chemicals}])>"


class IdealPoyintingCorrectionFactors(PoyintingCorrectionFactors):
    """Create an IdealPoyintingCorrectionFactor object that estimates all poyinting correction factors to be 1 when called with composition and temperature (K).
    
    Parameters
    ----------
    
    chemicals : Iterable[:class:`~thermosteam.Chemical`]
    
    """
    __slots__ = ('_chemicals')
    
    @property
    def chemicals(self):
        return self._chemicals
    @chemicals.setter
    def chemicals(self, chemicals):
        self._chemicals = tuple(chemicals)

    def __call__(self, y, T):
        return 1.



# -*- coding: utf-8 -*-
# BioSTEAM: The Biorefinery Simulation and Techno-Economic Analysis Modules
# Copyright (C) 2020, Yoel Cortes-Pena <yoelcortes@gmail.com>
# 
# This module is under the UIUC open-source license. See 
# github.com/BioSTEAMDevelopmentGroup/biosteam/blob/master/LICENSE.txt
# for license details.
"""
"""
from . import chemicals_user
from . import thermo_user
from . import read_only
from . import registered
from . import forward

__all__ = (*chemicals_user.__all__,
           *thermo_user.__all__,
           *read_only.__all__,
           *registered.__all__,
           *forward.__all__)

from .registered import *
from .chemicals_user import *
from .thermo_user import *
from .read_only import *
from .forward import *


# -*- coding: utf-8 -*-
# BioSTEAM: The Biorefinery Simulation and Techno-Economic Analysis Modules
# Copyright (C) 2020, Yoel Cortes-Pena <yoelcortes@gmail.com>
# 
# This module extends the vapor_pressure module from the chemicals's library:
# https://github.com/CalebBell/chemicals
# Copyright (C) 2020 Caleb Bell <Caleb.Andrew.Bell@gmail.com>
#
# This module is under a dual license:
# 1. The UIUC open-source license. See 
# github.com/BioSTEAMDevelopmentGroup/biosteam/blob/master/LICENSE.txt
# for license details.
# 
# 2. The MIT open-source license. See
# https://github.com/CalebBell/chemicals/blob/master/LICENSE.txt for details.
from chemicals import vapor_pressure as vp
from ..base import functor, TDependentHandleBuilder
from chemicals.dippr import EQ101
import numpy as np
from .data import (
    Psat_data_WagnerMcGarry,
    Psat_data_WagnerPoling,
    Psat_data_AntoinePoling,
    Psat_data_AntoineExtended,
    Psat_data_Perrys2_8,
    Psat_data_VDI_PPDS_3,
)

# %% Vapor pressure

vp.__all__.extend([
    'vapor_pressure_handle',
])

Antoine = functor(vp.Antoine, 'Psat')
TRC_Antoine_extended = functor(vp.TRC_Antoine_extended, 'Psat')
Wagner_original = functor(vp.Wagner_original, 'Psat')
Wagner = functor(vp.Wagner, 'Psat')
boiling_critical_relation = functor(vp.boiling_critical_relation , 'Psat')
Lee_Kesler = functor(vp.Lee_Kesler, 'Psat')
Ambrose_Walton = functor(vp.Ambrose_Walton, 'Psat')
Sanjari = functor(vp.Sanjari, 'Psat')
Edalat = functor(vp.Edalat, 'Psat')

@TDependentHandleBuilder('Psat')
def vapor_pressure_handle(handle, CAS, Tb, Tc, Pc, omega):
    add_model = handle.add_model
    if CAS in Psat_data_WagnerMcGarry:
        a, b, c, d, Pc, Tc, Tmin = Psat_data_WagnerMcGarry[CAS]
        Tmax = Tc
        data = (Tc, Pc, a, b, c, d)
        add_model(Wagner_original.functor.from_args(data), Tmin, Tmax)
    if CAS in Psat_data_WagnerPoling:
        a, b, c, d, Tc, Pc, Tmin, Tmax = Psat_data_WagnerPoling[CAS]
        # Some Tmin values are missing; Arbitrary choice of 0.1 lower limit
        if np.isnan(Tmin): Tmin = Tmax * 0.1
        data = (Tc, Pc, a, b, c, d)
        add_model(Wagner_original.functor.from_args(data), Tmin, Tmax)
    if CAS in Psat_data_AntoineExtended:
        a, b, c, Tc, to, n, e, f, Tmin, Tmax = Psat_data_AntoineExtended[CAS]
        data = (Tc, to, a, b, c, n, e, f)
        add_model(TRC_Antoine_extended.functor.from_args(data), Tmin, Tmax)
    if CAS in Psat_data_AntoinePoling:
        a, b, c, Tmin, Tmax = Psat_data_AntoinePoling[CAS]
        data = (a, b, c)
        add_model(Antoine.functor.from_args(data), Tmin, Tmax)
    if CAS in Psat_data_Perrys2_8:
        C1, C2, C3, C4, C5, Tmin, Tmax = Psat_data_Perrys2_8[CAS]
        data = (C1, C2, C3, C4, C5)
        add_model(EQ101.functor.from_args(data), Tmin, Tmax)
    if CAS in Psat_data_VDI_PPDS_3:
        Tm, Tc, Pc, a, b, c, d = Psat_data_VDI_PPDS_3[CAS]
        data = (Tc, Pc, a, b, c, d)
        add_model(Wagner.functor.from_args(data), 0., Tc,)
    data = (Tb, Tc, Pc)
    if all(data):
        add_model(boiling_critical_relation.functor.from_args(data), 0., Tc)
    data = (Tc, Pc, omega)
    if all(data):
        for f in (Lee_Kesler, Ambrose_Walton, Sanjari, Edalat):
            add_model(f.functor.from_args(data), 0., Tc)
vp.vapor_pressure_handle = vapor_pressure_handle


# -*- coding: utf-8 -*-
# BioSTEAM: The Biorefinery Simulation and Techno-Economic Analysis Modules
# Copyright (C) 2020, Yoel Cortes-Pena <yoelcortes@gmail.com>
# 
# This module is under the UIUC open-source license. See 
# github.com/BioSTEAMDevelopmentGroup/biosteam/blob/master/LICENSE.txt
# for license details.
"""
"""
from numpy import asarray, array
import flexsolve as flx
from .. import functional as fn
from ..exceptions import DomainError, InfeasibleRegion
from .solve_vle_composition import solve_x
from ..utils import fill_like, Cache
from .._settings import settings

__all__ = ('DewPoint', 'DewPointCache')

# %% Dew point values container

class DewPointValues:
    __slots__ = ('T', 'P', 'IDs', 'z', 'x')
    
    def __init__(self, T, P, IDs, z, x):
        self.T = T
        self.P = P
        self.IDs = IDs
        self.z = z
        self.x = x
        
    def __repr__(self):
        return f"{type(self).__name__}(T={self.T:.2f}, P={self.P:.0f}, IDs={self.IDs}, z={self.z}, x={self.x})"


# %% Dew point calculation

class DewPoint:
    """
    Create a DewPoint object that returns dew point values when called with a 
    composition and either a temperture (T) or pressure (P).
    
    Parameters
    ----------
    chemicals=None : Iterable[:class:`~thermosteam.Chemical`], optional
    
    thermo=None : :class:`~thermosteam.Thermo`, optional
    
    Examples
    --------
    >>> import thermosteam as tmo
    >>> chemicals = tmo.Chemicals(['Water', 'Ethanol'], cache=True)
    >>> tmo.settings.set_thermo(chemicals)
    >>> DP = tmo.equilibrium.DewPoint(chemicals)
    >>> # Solve for dew point at constant temperautre
    >>> molar_composition = (0.5, 0.5)
    >>> dp = DP(z=molar_composition, T=355)
    >>> dp
    DewPointValues(T=355.00, P=91970, IDs=('Water', 'Ethanol'), z=[0.5 0.5], x=[0.851 0.149])
    >>> # Note that the result is a DewPointValues object which contain all results as attibutes
    >>> (dp.T, round(dp.P), dp.IDs, dp.z, dp.x)
    (355, 91970, ('Water', 'Ethanol'), array([0.5, 0.5]), array([0.851, 0.149]))
    >>> # Solve for dew point at constant pressure
    >>> DP(z=molar_composition, P=2*101324)
    DewPointValues(T=376.26, P=202648, IDs=('Water', 'Ethanol'), z=[0.5 0.5], x=[0.832 0.168])

    """
    __slots__ = ('chemicals', 'phi', 'gamma', 'IDs',
                 'pcf', 'Psats', 'P', 'T', 'x',
                 'Tmin', 'Tmax', 'Pmin', 'Pmax')
    Tmin_default = 150.
    _cached = {}
    def __init__(self, chemicals=(), thermo=None):
        thermo = settings.get_default_thermo(thermo)
        chemicals = tuple(chemicals)
        key = (chemicals, thermo.Gamma, thermo.Phi, thermo.PCF)
        cached = self._cached
        if key in cached:
            other = cached[key]
            fill_like(self, other, self.__slots__)
        else:
            self.IDs = tuple([i.ID for i in chemicals])
            self.gamma = thermo.Gamma(chemicals)
            self.phi = thermo.Phi(chemicals)
            self.pcf = thermo.PCF(chemicals)
            self.Psats = Psats = [i.Psat for i in chemicals]
            self.Tmin = Tmin = max(max([i.Tmin for i in Psats]) + 1e-3, self.Tmin_default)
            self.Tmax = Tmax = min([i.Tmax for i in Psats]) - 1e-3
            self.Pmin = min([i(Tmin) for i in Psats])
            self.Pmax = max([i(Tmax) for i in Psats])
            self.chemicals = chemicals
            self.P = self.T = self.x = None
            cached[key] = self
    
    def _T_error(self, T, P, z_norm, zP):
        if T <= 0: raise InfeasibleRegion('negative temperature')
        Psats = array([i(T) for i in self.Psats])
        Psats[Psats < 1e-16] = 1e-16 # Prevent floating point error
        phi = self.phi(z_norm, T, P)
        x_gamma_pcf = phi * zP / Psats
        self.x = solve_x(x_gamma_pcf, self.gamma, self.pcf, T, self.x)
        return 1 - self.x.sum()
    
    def _T_error_ideal(self, T, zP):
        Psats = array([i(T) for i in self.Psats])
        Psats[Psats < 1e-16] = 1e-16 # Prevent floating point error
        self.x = zP / Psats
        return 1 - self.x.sum()
    
    def _P_error(self, P, T, z_norm, z_over_Psats):
        if P <= 0: raise InfeasibleRegion('negative pressure')
        x_gamma_pcf = z_over_Psats * P * self.phi(z_norm, T, P)
        self.x = solve_x(x_gamma_pcf, self.gamma, self.pcf, T, self.x)
        return 1 - self.x.sum()
    
    def _T_ideal(self, zP):
        f = self._T_error_ideal
        Tmin = self.Tmin
        Tmax = self.Tmax
        args = (zP,)
        T = flx.IQ_interpolation(f, Tmin, Tmax,
                                 f(Tmin, *args), f(Tmax, *args),
                                 None, 1e-9, 5e-12, args,
                                 checkiter=False, checkbounds=False)
        return T
    
    def _P_ideal(self, z_over_Psats):
        return 1. / z_over_Psats.sum()
    
    def __call__(self, z, *, T=None, P=None):
        z = asarray(z, float)
        if T:
            if P: raise ValueError("may specify either T or P, not both")
            P, x = self.solve_Px(z, T)
        elif P:
            T, x = self.solve_Tx(z, P)
        else:
            raise ValueError("must specify either T or P")
        return DewPointValues(T, P, self.IDs, z, x)
    
    def solve_Tx(self, z, P):
        """
        Dew point given composition and pressure.

        Parameters
        ----------
        z : ndarray
            Molar composition.
        P : float
            Pressure [Pa].

        Returns
        -------
        T : float
            Dew point temperature [K].
        x : numpy.ndarray
            Liquid phase molar composition.

        Examples
        --------
        >>> import thermosteam as tmo
        >>> import numpy as np
        >>> chemicals = tmo.Chemicals(['Water', 'Ethanol'], cache=True)
        >>> tmo.settings.set_thermo(chemicals)
        >>> DP = tmo.equilibrium.DewPoint(chemicals)
        >>> DP.solve_Tx(z=np.array([0.5, 0.5]), P=101325)
        (357.451847, array([0.849, 0.151]))
        """
        f = self._T_error
        z_norm = z/z.sum()
        zP = z * P
        args = (P, z_norm, zP)
        self.P = P
        T_guess = self.T or self._T_ideal(zP) 
        try:
            T = flx.aitken_secant(f, T_guess, T_guess + 1e-3,
                                  1e-9, 5e-12, args,
                                  checkiter=False)
        except (InfeasibleRegion, DomainError):
            Tmin = self.Tmin
            Tmax = self.Tmax
            T = flx.IQ_interpolation(f, Tmin, Tmax,
                                     f(Tmin, *args), f(Tmax, *args),
                                     T_guess, 1e-9, 5e-12, args,
                                     checkiter=False, checkbounds=False)
        self.x = fn.normalize(self.x)
        return T, self.x.copy()
    
    def solve_Px(self, z, T):
        """
        Dew point given composition and temperature.

        Parameters
        ----------
        z : ndarray
            Molar composition.
        T : float
            Temperature (K).
        
        Returns
        -------
        P : float
            Dew point pressure (Pa).
        x : ndarray
            Liquid phase molar composition.

        Examples
        --------
        >>> import thermosteam as tmo
        >>> import numpy as np
        >>> chemicals = tmo.Chemicals(['Water', 'Ethanol'], cache=True)
        >>> tmo.settings.set_thermo(chemicals)
        >>> DP = tmo.equilibrium.DewPoint(chemicals)
        >>> DP.solve_Px(z=np.array([0.5, 0.5]), T=352.28)
        (82444.29876, array([0.853, 0.147]))
 
       """
        z_norm = z/z.sum()
        Psats = array([i(T) for i in self.Psats], dtype=float)
        z_over_Psats = z/Psats
        args = (T, z_norm, z_over_Psats)
        self.T = T
        f = self._P_error
        P_guess = self.P or self._P_ideal(z_over_Psats)
        try:
            P = flx.aitken_secant(f, P_guess, P_guess-10, 1e-3, 5e-12, args,
                                  checkiter=False)
        except (InfeasibleRegion, DomainError):
            Pmin = self.Pmin
            Pmax = self.Pmax
            P = flx.IQ_interpolation(f, Pmin, Pmax, 
                                     f(Pmin, *args), f(Pmax, *args),
                                     P_guess, 1e-3, 5e-12, args,
                                     checkiter=False, checkbounds=False)
        self.x = fn.normalize(self.x)
        return P, self.x.copy()
    
    def __repr__(self):
        chemicals = ", ".join([i.ID for i in self.chemicals])
        return f"{type(self).__name__}([{chemicals}])"
    
class DewPointCache(Cache): load = DewPoint
del Cache


# -*- coding: utf-8 -*-
# BioSTEAM: The Biorefinery Simulation and Techno-Economic Analysis Modules
# Copyright (C) 2020, Yoel Cortes-Pena <yoelcortes@gmail.com>
# 
# This module is under the UIUC open-source license. See 
# github.com/BioSTEAMDevelopmentGroup/biosteam/blob/master/LICENSE.txt
# for license details.
"""
"""
import os
import json

# %% Readers

class RowData:
    """
    Create a RowData object for fast data frame lookups by row.
    
    """
    __slots__ = ('index', 'values', 'columns')
    
    def __init__(self, index, values=None, columns=None):
        self.index = {key: i for i, key in enumerate(index)}
        self.values = values
        self.columns = columns if columns is None else {key: i for i, key in enumerate(columns)} 
    
    def get(self, row, col):
        columns = self.columns
        if not columns: raise RuntimeError('no columns implemented')
        return self.values[self.index[row], columns[col]]
    
    def __getitem__(self, row):
        return self.values[self.index[row]]
    
    def __contains__(self, row):
        return row in self.index   
    
def load_json(folder, json_file, hook=None):
    with open(os.path.join(folder, json_file)) as f:
        return json.loads(f.read(), object_hook=hook)
    
# %% Data

### Volume ###

from chemicals.volume import (
    rho_data_COSTALD, 
    rho_data_SNM0,
    rho_data_Perry_8E_105_l, rho_values_Perry_8E_105_l,
    rho_data_VDI_PPDS_2, rho_values_VDI_PPDS_2,
    rho_data_CRC_inorg_l, rho_values_CRC_inorg_l,
    rho_data_CRC_inorg_l_const,
    rho_data_CRC_inorg_s_const, 
    rho_data_CRC_virial, rho_values_CRC_virial,
)
from chemicals.miscdata import (
    VDI_saturation_dict,
    lookup_VDI_tabular_data,    
)

rho_data_COSTALD = RowData(rho_data_COSTALD.index, 
                           rho_data_COSTALD.values,
                           rho_data_COSTALD.columns)
rho_data_SNM0 = RowData(rho_data_SNM0.index, 
                        rho_data_SNM0.values,
                        rho_data_SNM0.columns)
rho_data_Perry_8E_105_l = RowData(rho_data_Perry_8E_105_l.index,
                                  rho_values_Perry_8E_105_l)
rho_data_VDI_PPDS_2 = RowData(rho_data_VDI_PPDS_2.index, 
                              rho_values_VDI_PPDS_2)
rho_data_CRC_inorg_l = RowData(rho_data_CRC_inorg_l.index, 
                               rho_values_CRC_inorg_l)
rho_data_CRC_inorg_l_const = RowData(rho_data_CRC_inorg_l_const.index, 
                                     rho_data_CRC_inorg_l_const.values,
                                     rho_data_CRC_inorg_l_const.columns)
rho_data_CRC_inorg_s_const = RowData(rho_data_CRC_inorg_s_const, 
                                     rho_data_CRC_inorg_s_const.values,
                                     rho_data_CRC_inorg_s_const.columns)
rho_data_CRC_virial = RowData(rho_data_CRC_virial.index, 
                              rho_values_CRC_virial)

### Phase change ###

from chemicals.phase_change import (
    phase_change_data_Perrys2_150, phase_change_values_Perrys2_150,
    phase_change_data_VDI_PPDS_4, phase_change_values_VDI_PPDS_4,
    phase_change_data_Alibakhshi_Cs, 
    Hvap_data_CRC,
    Hvap_data_Gharagheizi,
)

phase_change_data_Perrys2_150 = RowData(phase_change_data_Perrys2_150.index,
                                        phase_change_values_Perrys2_150)
phase_change_data_VDI_PPDS_4 = RowData(phase_change_data_VDI_PPDS_4.index, 
                                       phase_change_values_VDI_PPDS_4)
phase_change_data_Alibakhshi_Cs = RowData(phase_change_data_Alibakhshi_Cs.index,
                                          phase_change_data_Alibakhshi_Cs.values,
                                          phase_change_data_Alibakhshi_Cs.columns)
Hvap_data_CRC = RowData(Hvap_data_CRC.index, 
                        Hvap_data_CRC.values,
                        Hvap_data_CRC.columns)
Hvap_data_Gharagheizi = RowData(Hvap_data_Gharagheizi.index, 
                                Hvap_data_Gharagheizi.values,
                                Hvap_data_Gharagheizi.columns)

### Heat capacity ###

from chemicals.heat_capacity import (
    Cp_data_Poling,
    Cp_values_Poling,
    TRC_gas_data,
    TRC_gas_values,
    CRC_standard_data,
    Cp_dict_PerryI,
    zabransky_dict_sat_s,
    zabransky_dict_sat_p,
    zabransky_dict_const_s,
    zabransky_dict_const_p,
    zabransky_dict_iso_s,
    zabransky_dict_iso_p,
)

Cp_data_Poling = RowData(Cp_data_Poling.index, Cp_values_Poling)
TRC_gas_data = RowData(TRC_gas_data.index, TRC_gas_values)
CRC_standard_data = RowData(CRC_standard_data)

### Permitivity ###

from chemicals.permittivity import (
    permittivity_data_CRC, permittivity_values_CRC
)

permittivity_data_CRC = RowData(permittivity_data_CRC.index,
                                permittivity_values_CRC)

### Surface tension ###

from chemicals.interface import (
    sigma_data_Mulero_Cachadina, sigma_values_Mulero_Cachadina,
    sigma_data_Jasper_Lange, sigma_values_Jasper_Lange,
    sigma_data_Somayajulu, sigma_values_Somayajulu, sigma_data_Somayajulu2,
    sigma_values_Somayajulu2, sigma_data_VDI_PPDS_11, sigma_values_VDI_PPDS_11
)

sigma_data_Mulero_Cachadina = RowData(sigma_data_Mulero_Cachadina.index, sigma_values_Mulero_Cachadina)
sigma_data_Jasper_Lange = RowData(sigma_data_Jasper_Lange.index, sigma_values_Jasper_Lange,)
sigma_data_Somayajulu = RowData(sigma_data_Somayajulu.index, sigma_values_Somayajulu)
sigma_data_Somayajulu2 = RowData(sigma_data_Somayajulu2.index, sigma_values_Somayajulu2)
sigma_data_VDI_PPDS_11 = RowData(sigma_data_VDI_PPDS_11.index, sigma_values_VDI_PPDS_11)

### Thermal conductivity ###

from chemicals.thermal_conductivity import (
    k_data_Perrys_8E_2_315, k_values_Perrys_8E_2_315, 
    k_data_VDI_PPDS_9, k_values_VDI_PPDS_9
)

k_data_Perrys_8E_2_315 = RowData(k_data_Perrys_8E_2_315.index, k_values_Perrys_8E_2_315)
k_data_VDI_PPDS_9 = RowData(k_data_VDI_PPDS_9.index, k_values_VDI_PPDS_9)

### Vapor pressure ###

from chemicals.vapor_pressure import (
    Psat_data_WagnerMcGarry, Psat_values_WagnerMcGarry,
    Psat_data_WagnerPoling, Psat_values_WagnerPoling,
    Psat_data_AntoinePoling, Psat_values_AntoinePoling,
    Psat_data_AntoineExtended, Psat_values_AntoineExtended,
    Psat_data_Perrys2_8, Psat_values_Perrys2_8,
    Psat_data_VDI_PPDS_3, Psat_values_VDI_PPDS_3
)

Psat_data_WagnerMcGarry = RowData(Psat_data_WagnerMcGarry.index, Psat_values_WagnerMcGarry)
Psat_data_WagnerPoling = RowData(Psat_data_WagnerPoling.index, Psat_values_WagnerPoling)
Psat_data_AntoinePoling = RowData(Psat_data_AntoinePoling.index, Psat_values_AntoinePoling)
Psat_data_AntoineExtended = RowData(Psat_data_AntoineExtended.index, Psat_values_AntoineExtended)
Psat_data_Perrys2_8 = RowData(Psat_data_Perrys2_8.index, Psat_values_Perrys2_8)
Psat_data_VDI_PPDS_3 = RowData(Psat_data_VDI_PPDS_3.index, Psat_values_VDI_PPDS_3)

### Viscosity ###

from chemicals.viscosity import (
    mu_data_Dutt_Prasad, mu_values_Dutt_Prasad,
    mu_data_VN3, mu_values_VN3,
    mu_data_VN2, mu_values_VN2,
    mu_data_VN2E, mu_values_VN2E, 
    mu_data_Perrys_8E_2_313, mu_values_Perrys_8E_2_313,
    mu_data_Perrys_8E_2_312, mu_values_Perrys_8E_2_312,
    mu_data_VDI_PPDS_7, mu_values_PPDS_7, 
    mu_data_VDI_PPDS_8, mu_values_PPDS_8,
)
mu_data_Dutt_Prasad = RowData(mu_data_Dutt_Prasad.index, mu_values_Dutt_Prasad)
mu_data_VN3 = RowData(mu_data_VN3.index, mu_values_VN3)
mu_data_VN2 = RowData(mu_data_VN2.index, mu_values_VN2)
mu_data_Perrys_8E_2_313 = RowData(mu_data_Perrys_8E_2_313.index, mu_values_Perrys_8E_2_313)
mu_data_Perrys_8E_2_312 = RowData(mu_data_Perrys_8E_2_312.index, mu_values_Perrys_8E_2_312)
mu_data_VDI_PPDS_7 = RowData(mu_data_VDI_PPDS_7.index, mu_values_PPDS_7)
mu_data_VDI_PPDS_8 = RowData(mu_data_VDI_PPDS_8.index, mu_values_PPDS_8)



# -*- coding: utf-8 -*-
# BioSTEAM: The Biorefinery Simulation and Techno-Economic Analysis Modules
# Copyright (C) 2020, Yoel Cortes-Pena <yoelcortes@gmail.com>
# 
# This module extends the dippr module from the chemicals's library:
# https://github.com/CalebBell/chemicals
# Copyright (C) 2020 Caleb Bell <Caleb.Andrew.Bell@gmail.com>
#
# This module is under a dual license:
# 1. The UIUC open-source license. See 
# github.com/BioSTEAMDevelopmentGroup/biosteam/blob/master/LICENSE.txt
# for license details.
# 
# 2. The MIT open-source license. See
# https://github.com/CalebBell/chemicals/blob/master/LICENSE.txt for details.
from chemicals import dippr
from ..base import functor

EQ100 = functor(dippr.EQ100)
EQ101 = functor(dippr.EQ101)
EQ102 = functor(dippr.EQ102)
EQ104 = functor(dippr.EQ104)
EQ105 = functor(dippr.EQ105)
EQ106 = functor(dippr.EQ106)
EQ107 = functor(dippr.EQ107)
EQ114 = functor(dippr.EQ114)
EQ115 = functor(dippr.EQ115)
EQ116 = functor(dippr.EQ116)
EQ127 = functor(dippr.EQ127)



# -*- coding: utf-8 -*-
# BioSTEAM: The Biorefinery Simulation and Techno-Economic Analysis Modules
# Copyright (C) 2020, Yoel Cortes-Pena <yoelcortes@gmail.com>
# 
# This module is under the UIUC open-source license. See 
# github.com/BioSTEAMDevelopmentGroup/biosteam/blob/master/LICENSE.txt
# for license details.
"""
"""
import flexsolve as flx

__all__ = ('Mixture',)

def iter_temperature(T, H, H_model, phase, mol, P, Cn):
    # Used to solve for ethalpy at given temperature
    return T + (H - H_model(phase, mol, T, P)) / Cn

def xiter_temperature(T, H, H_model, phase_mol, P, Cn):
    # Used to solve for ethalpy at given temperature
    return T + (H - H_model(phase_mol, T, P)) / Cn

# %% Ideal mixture

class Mixture:
    """
    Create an Mixture object for estimating mixture properties.
    
    Parameters
    ----------
    rule : str
        Description of mixing rules used.
    Cn : function(phase, mol, T)
        Molar heat capacity mixture model [J/mol/K].
    H : function(phase, mol, T)
        Enthalpy mixture model [J/mol].
    S : function(phase, mol, T, P)
        Entropy mixture model [J/mol].
    H_excess : function(phase, mol, T, P)
        Excess enthalpy mixture model [J/mol].
    S_excess : function(phase, mol, T, P)
        Excess entropy mixture model [J/mol].
    mu : function(phase, mol, T, P)
        Dynamic viscosity mixture model [Pa*s].
    V : function(phase, mol, T, P)
        Molar volume mixture model [m^3/mol].
    kappa : function(phase, mol, T, P)
        Thermal conductivity mixture model [W/m/K].
    Hvap : function(mol, T)
        Heat of vaporization mixture model [J/mol]
    sigma : function(mol, T, P)
        Surface tension mixture model [N/m].
    epsilon : function(mol, T, P)
        Relative permitivity mixture model [-]
    include_excess_energies=False : bool
        Whether to include excess energies
        in enthalpy and entropy calculations.
    
    Notes
    -----
    Although the mixture models are on a molar basis, this is only if the molar
    data is normalized before the calculation (i.e. the `mol` parameter is 
    normalized before being passed to the model).
    
    See also
    --------
    IdealMixtureModel
    :func:`~.mixture_builders.ideal_mixture`
    
    Attributes
    ----------
    rule : str
        Description of mixing rules used.
    include_excess_energies : bool
        Whether to include excess energies
        in enthalpy and entropy calculations.
    Cn(phase, mol, T) : 
        Mixture molar heat capacity [J/mol/K].
    mu(phase, mol, T, P) : 
        Mixture dynamic viscosity [Pa*s].
    V(phase, mol, T, P) : 
        Mixture molar volume [m^3/mol].
    kappa(phase, mol, T, P) : 
        Mixture thermal conductivity [W/m/K].
    Hvap(mol, T, P) : 
        Mixture heat of vaporization [J/mol]
    sigma(mol, T, P) : 
        Mixture surface tension [N/m].
    epsilon(mol, T, P) : 
        Mixture relative permitivity [-].
    
    
    """
    __slots__ = ('rule',
                 'rigorous_energy_balance',
                 'include_excess_energies',
                 'Cn', 'mu', 'V', 'kappa',
                 'Hvap', 'sigma', 'epsilon',
                 '_H', '_H_excess', '_S', '_S_excess',
    )
    
    def __init__(self, rule, Cn, H, S, H_excess, S_excess,
                 mu, V, kappa, Hvap, sigma, epsilon,
                 include_excess_energies=False):
        self.rule = rule
        self.include_excess_energies = include_excess_energies
        self.Cn = Cn
        self.mu = mu
        self.V = V
        self.kappa = kappa
        self.Hvap = Hvap
        self.sigma = sigma
        self.epsilon = epsilon
        self._H = H
        self._S = S
        self._H_excess = H_excess
        self._S_excess = S_excess
    
    def H(self, phase, mol, T, P):
        """Return enthalpy [J/mol]."""
        H = self._H(phase, mol, T)
        if self.include_excess_energies:
            H += self._H_excess(phase, mol, T, P)
        return H
    
    def S(self, phase, mol, T, P):
        """Return entropy in [J/mol]."""
        S = self._S(phase, mol, T, P)
        if self.include_excess_energies:
            S += self._S_excess(phase, mol, T, P)
        return S
    
    def solve_T(self, phase, mol, H, T_guess, P):
        """Solve for temperature in Kelvin."""
        args = (H, self.H, phase, mol, P, self.Cn(phase, mol, T_guess))
        return flx.aitken(iter_temperature, T_guess, 1e-6, args, 50, checkiter=True)
        
    def xsolve_T(self, phase_mol, H, T_guess, P):
        """Solve for temperature in Kelvin."""
        args = (H, self.xH, tuple(phase_mol), P, self.xCn(phase_mol, T_guess))
        return flx.aitken(xiter_temperature, T_guess, 1e-6, args, 50, checkiter=True)
        
    def xCn(self, phase_mol, T):
        """Multi-phase mixture heat capacity [J/mol/K]."""
        Cn = self.Cn
        return sum([Cn(phase, mol, T) for phase, mol in phase_mol])
    
    def xH(self, phase_mol, T, P):
        """Multi-phase mixture enthalpy [J/mol]."""
        H = self._H
        H_total = sum([H(phase, mol, T) for phase, mol in phase_mol])
        if self.include_excess_energies:
            H_excess = self._H_excess
            H_total += sum([H_excess(phase, mol, T, P) for phase, mol in phase_mol])
        return H_total
    
    def xS(self, phase_mol, T, P):
        """Multi-phase mixture entropy [J/mol]."""
        S = self._S
        S_total = sum([S(phase, mol, T, P) for phase, mol in phase_mol])
        if self.include_excess_energies:
            S_excess = self._S_excess
            S_total += sum([S_excess(phase, mol, T, P) for phase, mol in phase_mol])
        return S_total
    
    def xV(self, phase_mol, T, P):
        """Multi-phase mixture molar volume [mol/m^3]."""
        V = self.V
        return sum([V(phase, mol, T, P) for phase, mol in phase_mol])
    
    def xmu(self, phase_mol, T, P):
        """Multi-phase mixture hydrolic [Pa*s]."""
        mu = self.mu
        return sum([mu(phase, mol, T, P) for phase, mol in phase_mol])
    
    def xkappa(self, phase_mol, T, P):
        """Multi-phase mixture thermal conductivity [W/m/K]."""
        kappa = self.kappa
        return sum([kappa(phase, mol, T, P) for phase, mol in phase_mol])
    
    def __repr__(self):
        return f"{type(self).__name__}(rule={repr(self.rule)}, ..., include_excess_energies={self.include_excess_energies})"
    
    def _info(self):
        return (f"{type(self).__name__}(\n"
                f"    rule={repr(self.rule)}, ...\n"
                f"    include_excess_energies={self.include_excess_energies}\n"
                 ")")
    
    def show(self):
        print(self._info())
    _ipython_display_ = show
        


# -*- coding: utf-8 -*-
# BioSTEAM: The Biorefinery Simulation and Techno-Economic Analysis Modules
# Copyright (C) 2020, Yoel Cortes-Pena <yoelcortes@gmail.com>
# 
# This module is under the UIUC open-source license. See 
# github.com/BioSTEAMDevelopmentGroup/biosteam/blob/master/LICENSE.txt
# for license details.
from distutils.core import setup

setup(
    name='thermosteam',
    packages=['thermosteam'],
    license='MIT',
    version='0.23.1',
    description="BioSTEAM's Premier Thermodynamic Engine",
    long_description=open('README.rst').read(),
    author='Yoel Cortes-Pena',
    install_requires=['pint>=0.9',
                      'scipy>=1.5', 
                      'chemicals>=0.1.5',
                      'IPython>=7.9.0', 
                      'colorpalette>=0.3.1', 
                      'biosteam>=2.21.11',
                      'pandas>=0.25.2', 
                      'matplotlib>=3.1.1',
                      'numpy>=1.18.1', 
                      'xlrd==1.2.0',
                      'openpyxl>=3.0.0', 
                      'free_properties>=0.2.5',
                      'flexsolve==0.3.32', 
                      'pyglet',
                      'numba>=0.50.0'],
    extras_require={ # Not in current PyPI version, but will be available in the next
        'dev': [
            'sympy',
            'sphinx', 
            'sphinx_rtd_theme', 
            'pyyaml',
            'pytest-cov',
            'coveralls',
        ]
    }, 
    package_data={
        'thermosteam': [
            'base/*', 
            'equilibrium/*', 
            'reaction/*',
            'utils/*',
            'utils/decorators/*', 
            'mixture/*',
            'chemicals/*',
            'chemicals/Reaction/*',
            'equilibrium/UNIFAC/*',
            'units_of_measure.txt', 
        ]
    },
    platforms=['Windows', 'Mac', 'Linux'],
    author_email='yoelcortes@gmail.com',
    url='https://github.com/BioSTEAMDevelopmentGroup/thermosteam',
    download_url='https://github.com/BioSTEAMDevelopmentGroup/thermosteam.git',
    classifiers=['Development Status :: 3 - Alpha',
                 'Environment :: Console',
                 'License :: OSI Approved :: University of Illinois/NCSA Open Source License',
                 'License :: OSI Approved :: MIT License',
                 'Topic :: Scientific/Engineering',
                 'Topic :: Scientific/Engineering :: Chemistry',
                 'Topic :: Scientific/Engineering :: Mathematics',
                 'Intended Audience :: Developers',
                 'Intended Audience :: Education',
                 'Intended Audience :: Manufacturing',
                 'Intended Audience :: Science/Research',
                 'Natural Language :: English',
                 'Operating System :: MacOS',
                 'Operating System :: Microsoft :: Windows',
                 'Operating System :: POSIX',
                 'Operating System :: POSIX :: BSD',
                 'Operating System :: POSIX :: Linux',
                 'Operating System :: Unix',
                 'Programming Language :: Python :: 3.6',
                 'Programming Language :: Python :: 3.7',
                 'Programming Language :: Python :: 3.8',
                 'Programming Language :: Python :: Implementation :: CPython',
                 'Topic :: Education'],
    keywords=['thermodynamics', 'chemical engineering', 'mass and energy balance', 'material properties', 'phase equilibrium'],
)


# -*- coding: utf-8 -*-
# BioSTEAM: The Biorefinery Simulation and Techno-Economic Analysis Modules
# Copyright (C) 2020, Yoel Cortes-Pena <yoelcortes@gmail.com>
# 
# This module is under the UIUC open-source license. See 
# github.com/BioSTEAMDevelopmentGroup/biosteam/blob/master/LICENSE.txt
# for license details.
"""
This module contains functions for modeling separations in unit operations.

"""
import thermosteam as tmo
import flexsolve as flx
import numpy as np
from .utils import thermo_user
from .exceptions import InfeasibleRegion
from .equilibrium import phase_fraction

__all__ = (
    'split', 'mix_and_split',
    'adjust_moisture_content', 
    'mix_and_split_with_moisture_content',
    'partition_coefficients',
    'vle_partition_coefficients',
    'lle_partition_coefficients',
    'partition', 'lle', 'vle', 
    'material_balance',
    'StageLLE',
    'MultiStageLLE',
    'single_component_flow_rates_for_multi_stage_lle_without_side_draws',
    'flow_rates_for_multi_stage_extration_without_side_draws',
    'chemical_splits'
)


# %% Mixing, splitting, and moisture content

CAS_water = '7732-18-5'

def mix_and_split_with_moisture_content(ins, retentate, permeate,
                                        split, moisture_content):
    """
    Run splitter mass and energy balance with mixing all input streams and 
    and ensuring retentate moisture content.
    
    Parameters
    ----------
    ins : Iterable[Stream]
        Inlet fluids with solids.
    retentate : Stream
    permeate : Stream
    split : array_like
        Component splits to the retentate.
    moisture_content : float
        Fraction of water in retentate.

    Examples
    --------
    >>> import thermosteam as tmo
    >>> Solids = tmo.Chemical('Solids', default=True, search_db=False, phase='s')
    >>> tmo.settings.set_thermo(['Water', Solids])
    >>> feed = tmo.Stream('feed', Water=100, Solids=10, units='kg/hr')
    >>> wash_water = tmo.Stream('wash_water', Water=10, units='kg/hr')
    >>> retentate = tmo.Stream('retentate')
    >>> permeate = tmo.Stream('permeate')
    >>> split = [0., 1.]
    >>> moisture_content = 0.5
    >>> tmo.separations.mix_and_split_with_moisture_content(
    ...     [feed, wash_water], retentate, permeate, split, moisture_content
    ... )
    >>> retentate.show(flow='kg/hr')
    Stream: retentate
     phase: 'l', T: 298.15 K, P: 101325 Pa
     flow (kg/hr): Water   10
                   Solids  10
    >>> permeate.show(flow='kg/hr')
    Stream: permeate
     phase: 'l', T: 298.15 K, P: 101325 Pa
     flow (kg/hr): Water  100

    """
    mix_and_split(ins, retentate, permeate, split)
    adjust_moisture_content(retentate, permeate, moisture_content)

def adjust_moisture_content(retentate, permeate, moisture_content):
    """
    Remove water from permate to adjust retentate moisture content.
    
    Parameters
    ----------
    retentate : Stream
    permeate : Stream
    moisture_content : float
        Fraction of water in retentate.

    Examples
    --------
    >>> import thermosteam as tmo
    >>> Solids = tmo.Chemical('Solids', default=True, search_db=False, phase='s')
    >>> tmo.settings.set_thermo(['Water', Solids])
    >>> retentate = tmo.Stream('retentate', Solids=20, units='kg/hr')
    >>> permeate = tmo.Stream('permeate', Water=50, Solids=0.1, units='kg/hr')
    >>> moisture_content = 0.5
    >>> tmo.separations.adjust_moisture_content(retentate, permeate, moisture_content)
    >>> retentate.show(flow='kg/hr')
    Stream: retentate
     phase: 'l', T: 298.15 K, P: 101325 Pa
     flow (kg/hr): Water   20
                   Solids  20
    >>> permeate.show(flow='kg/hr')
    Stream: permeate
     phase: 'l', T: 298.15 K, P: 101325 Pa
     flow (kg/hr): Water   30
                   Solids  0.1
    
    Note that if not enough water is available, an InfeasibleRegion error is raised:
        
    >>> retentate.imol['Water'] = permeate.imol['Water'] = 0
    >>> tmo.separations.adjust_moisture_content(retentate, permeate, moisture_content)
    Traceback (most recent call last):
    InfeasibleRegion: not enough water; permeate moisture content is infeasible

    """
    F_mass = retentate.F_mass
    mc = moisture_content
    retentate_water = retentate.imol[CAS_water]
    dry_mass = F_mass - 18.01528 * retentate_water
    retentate.imol[CAS_water] = water = (dry_mass * mc/(1-mc))/18.01528
    permeate.imol[CAS_water] -= (water - retentate_water)
    if permeate.imol[CAS_water] < 0:
        raise InfeasibleRegion('not enough water; permeate moisture content')

def mix_and_split(ins, top, bottom, split):
    """
    Run splitter mass and energy balance with mixing all input streams.
    
    Parameters
    ----------
    ins : Iterable[Stream]
        All inlet fluids.
    top : Stream
        Top inlet fluid.
    bottom : Stream
        Bottom inlet fluid
    split : array_like
        Component-wise split of feed to the top stream.
    
    Examples
    --------
    >>> import thermosteam as tmo
    >>> tmo.settings.set_thermo(['Water', 'Ethanol'], cache=True)
    >>> feed_a = tmo.Stream(Water=20, Ethanol=5)
    >>> feed_b = tmo.Stream(Water=15, Ethanol=5)
    >>> split = 0.8
    >>> effluent_a = tmo.Stream('effluent_a')
    >>> effluent_b = tmo.Stream('effluent_b')
    >>> tmo.separations.mix_and_split([feed_a, feed_b], effluent_a, effluent_b, split)
    >>> effluent_a.show()
    Stream: effluent_a
     phase: 'l', T: 298.15 K, P: 101325 Pa
     flow (kmol/hr): Water    28
                     Ethanol  8
    >>> effluent_b.show()
    Stream: effluent_b
     phase: 'l', T: 298.15 K, P: 101325 Pa
     flow (kmol/hr): Water    7
                     Ethanol  2
    
    """
    top.mix_from(ins)
    bottom.copy_like(top)
    top_mol = top.mol
    top_mol[:] *= split
    bottom.mol[:] -= top_mol

def split(feed, top, bottom, split):
    """
    Run splitter mass and energy balance with mixing all input streams.
    
    Parameters
    ----------
    feed : Stream
        Inlet fluid.
    top : Stream
        Top inlet fluid.
    bottom : Stream
        Bottom inlet fluid
    split : array_like
        Component-wise split of feed to the top stream.
    
    Examples
    --------
    >>> import thermosteam as tmo
    >>> tmo.settings.set_thermo(['Water', 'Ethanol'], cache=True)
    >>> feed = tmo.Stream(Water=35, Ethanol=10)
    >>> split = 0.8
    >>> effluent_a = tmo.Stream('effluent_a')
    >>> effluent_b = tmo.Stream('effluent_b')
    >>> tmo.separations.split(feed, effluent_a, effluent_b, split)
    >>> effluent_a.show()
    Stream: effluent_a
     phase: 'l', T: 298.15 K, P: 101325 Pa
     flow (kmol/hr): Water    28
                     Ethanol  8
    >>> effluent_b.show()
    Stream: effluent_b
     phase: 'l', T: 298.15 K, P: 101325 Pa
     flow (kmol/hr): Water    7
                     Ethanol  2
                     
    """
    if feed is not top: top.copy_like(feed)
    bottom.copy_like(top)
    top_mol = top.mol
    top_mol[:] *= split
    bottom.mol[:] -= top_mol

def phase_split(feed, outlets):
    """
    Split the feed to outlets by phase.
    
    Parameters
    ----------
    feed : stream
    outlets : streams
        
    Notes
    -----
    Phases allocate to outlets in alphabetical order. For example,
    if the feed.phases is 'gls' (i.e. gas, liquid, and solid), the phases
    of the outlets will be 'g', 'l', and 's'.
        
    Examples
    --------
    Split gas and liquid phases to streams:
    
    >>> import thermosteam as tmo
    >>> tmo.settings.set_thermo(['Water', 'Ethanol'], cache=True)
    >>> feed = tmo.Stream('feed', Water=10, Ethanol=10)
    >>> feed.vle(V=0.5, P=101325)
    >>> vapor = tmo.Stream('vapor')
    >>> liquid = tmo.Stream('liquid')
    >>> outlets = [vapor, liquid]
    >>> tmo.separations.phase_split(feed, outlets)
    >>> vapor.show()
    Stream: vapor
     phase: 'g', T: 353.88 K, P: 101325 Pa
     flow (kmol/hr): Water    3.86
                     Ethanol  6.14
    >>> liquid.show()
    Stream: liquid
     phase: 'l', T: 353.88 K, P: 101325 Pa
     flow (kmol/hr): Water    6.14
                     Ethanol  3.86
    
    Note that the number of phases in the feed should be equal to the number of 
    outlets:
        
    >>> tmo.separations.phase_split(feed, [vapor])
    Traceback (most recent call last):
    RuntimeError: number of phases in feed must be equal to the number of outlets
    
    """
    phases = feed.phases
    if len(outlets) != len(phases):
        raise RuntimeError('number of phases in feed must be equal to the number of outlets')
    for i,j in zip(feed, outlets): j.copy_like(i)

# %% Single stage equilibrium

def partition_coefficients(IDs, top, bottom):
    """
    Return partition coefficients given streams in equilibrium.
    
    Parameters
    ----------
    top : Stream
        Vapor fluid.
    bottom : Stream
        Liquid fluid.
    IDs : tuple[str]
        IDs of chemicals in equilibrium.
    
    Returns
    -------
    K : 1d array
        Patition coefficients in mol fraction in top stream over mol fraction in bottom stream.

    Examples
    --------
    >>> import thermosteam as tmo
    >>> tmo.settings.set_thermo(['Water', 'Ethanol', tmo.Chemical('O2', phase='g')], cache=True)
    >>> s = tmo.Stream('s', Water=20, Ethanol=20, O2=0.1)
    >>> s.vle(V=0.5, P=101325)
    >>> tmo.separations.partition_coefficients(('Water', 'Ethanol'), s['g'], s['l'])
    array([0.629, 1.59 ])

    """
    return top.get_normalized_mol(IDs) / bottom.get_normalized_mol(IDs)

def chemical_splits(a, b=None, mixed=None):
    """
    Return a ChemicalIndexer with splits for all chemicals to stream `a`.
    
    Examples
    --------
    >>> import thermosteam as tmo
    >>> tmo.settings.set_thermo(['Water', 'Ethanol'], cache=True)
    >>> stream = tmo.Stream('stream', Water=10, Ethanol=10)
    >>> stream.vle(V=0.5, P=101325)
    >>> isplits = tmo.separations.chemical_splits(stream['g'], stream['l'])
    >>> isplits.show()
    ChemicalIndexer:
     Water    0.3861
     Ethanol  0.6139
    >>> isplits = tmo.separations.chemical_splits(stream['g'], mixed=stream)
    >>> isplits.show()
    ChemicalIndexer:
     Water    0.3861
     Ethanol  0.6139
     
    """
    mixed_mol = mixed.mol.copy() if mixed else a.mol + b.mol
    mixed_mol[mixed_mol==0.] = 1.
    return tmo.indexer.ChemicalIndexer.from_data(a.mol / mixed_mol)

def vle_partition_coefficients(top, bottom):
    """
    Return VLE partition coefficients given vapor and liquid streams in equilibrium.
    
    Parameters
    ----------
    top : Stream
        Vapor fluid.
    bottom : Stream
        Liquid fluid.
    
    Returns
    -------
    IDs : tuple[str]
        IDs for chemicals in vapor-liquid equilibrium.
    K : 1d array
        Patition coefficients in mol fraction in vapor over mol fraction in liquid.

    Examples
    --------
    >>> import thermosteam as tmo
    >>> tmo.settings.set_thermo(['Water', 'Ethanol', tmo.Chemical('O2', phase='g')], cache=True)
    >>> s = tmo.Stream('s', Water=20, Ethanol=20, O2=0.1)
    >>> s.vle(V=0.5, P=101325)
    >>> IDs, K = tmo.separations.vle_partition_coefficients(s['g'], s['l'])
    >>> IDs
    ('Water', 'Ethanol')
    >>> K
    array([0.629, 1.59 ])

    """
    IDs = tuple([i.ID for i in bottom.vle_chemicals])
    return IDs, partition_coefficients(IDs, top, bottom)

def lle_partition_coefficients(top, bottom):
    """
    Return LLE partition coefficients given two liquid streams in equilibrium.
    
    Parameters
    ----------
    top : Stream
        Liquid fluid.
    bottom : Stream
        Other liquid fluid.
    
    Returns
    -------
    IDs : tuple[str]
        IDs for chemicals in liquid-liquid equilibrium.
    K : 1d array
        Patition coefficients in mol fraction in top liquid over mol fraction in bottom liquid.

    Examples
    --------
    >>> import thermosteam as tmo
    >>> tmo.settings.set_thermo(['Water', 'Ethanol', 'Octanol'], cache=True)
    >>> s = tmo.Stream('s', Water=20, Octanol=20, Ethanol=1)
    >>> s.lle(T=298.15, P=101325)
    >>> IDs, K = tmo.separations.lle_partition_coefficients(s['l'], s['L'])
    >>> IDs
    ('Water', 'Ethanol', 'Octanol')
    >>> K
    array([6.82e+00, 2.38e-01, 3.00e-04])

    """
    IDs = tuple([i.ID for i in bottom.lle_chemicals])
    return IDs, partition_coefficients(IDs, top, bottom)

def partition(feed, top, bottom, IDs, K, phi=None):
    """
    Run equilibrium of feed to top and bottom streams given partition 
    coeffiecients and return the phase fraction.

    Parameters
    ----------
    feed : Stream
        Mixed feed.
    top : Stream
        Top fluid.
    bottom : Stream
        Bottom fluid.
    IDs : tuple[str]
        IDs of chemicals in equilibrium.
    K : 1d array
        Partition coefficeints corresponding to IDs.
    phi : float, optional
        Guess phase fraction in top phase.

    Returns
    -------
    phi : float
        Phase fraction in top phase.

    Notes
    -----
    Chemicals not in equilibrium end up in the top phase.

    Examples
    --------
    >>> import numpy as np
    >>> import thermosteam as tmo
    >>> tmo.settings.set_thermo(['Water', 'Ethanol', tmo.Chemical('O2', phase='g')], cache=True)
    >>> IDs = ('Water', 'Ethanol')
    >>> K = np.array([0.629, 1.59])
    >>> feed = tmo.Stream('feed', Water=20, Ethanol=20, O2=0.1)
    >>> top = tmo.Stream('top')
    >>> bottom = tmo.Stream('bottom')
    >>> tmo.separations.partition(feed, top, bottom, IDs, K)
    0.5002512677600628
    >>> top.show()
    Stream: top
     phase: 'l', T: 298.15 K, P: 101325 Pa
     flow (kmol/hr): Water    7.73
                     Ethanol  12.3
                     O2       0.1
    >>> bottom.show()
    Stream: bottom
     phase: 'l', T: 298.15 K, P: 101325 Pa
     flow (kmol/hr): Water    12.3
                     Ethanol  7.72

    """
    feed_mol = feed.mol
    mol = feed.imol[IDs]
    F_mol = mol.sum()
    z_mol = mol / F_mol
    phi = phase_fraction(z_mol, K, phi)
    x = z_mol / (phi * K + (1 - phi))
    bottom.empty()
    bottom.imol[IDs] = x * (1 - phi) * F_mol
    top.mol[:] = feed_mol - bottom.mol
    return phi

def lle(feed, top, bottom, top_chemical=None, efficiency=1.0, multi_stream=None):
    """
    Run LLE mass and energy balance.
    
    Parameters
    ----------
    feed : Stream
        Mixed feed.
    top : Stream
        Top fluid.
    bottom : Stream
        Bottom fluid.
    top_chemical : str, optional
        Identifier of chemical that will be favored in the top fluid.
    efficiency=1. : float,
        Fraction of feed in liquid-liquid equilibrium.
        The rest of the feed is divided equally between phases
    multi_stream : MultiStream, optional
        Data from feed is passed to this stream to perform liquid-liquid equilibrium.
    
    Examples
    --------
    Perform liquid-liquid equilibrium around water and octanol and split the phases:
    
    >>> import thermosteam as tmo
    >>> tmo.settings.set_thermo(['Water', 'Ethanol', 'Octanol'], cache=True)
    >>> feed = tmo.Stream('feed', Water=20, Octanol=20, Ethanol=1)
    >>> top = tmo.Stream('top')
    >>> bottom = tmo.Stream('bottom')
    >>> tmo.separations.lle(feed, top, bottom)
    >>> top.show()
    Stream: top
     phase: 'l', T: 298.15 K, P: 101325 Pa
     flow (kmol/hr): Water    3.55
                     Ethanol  0.861
                     Octanol  20
    >>> bottom.show()
    Stream: bottom
     phase: 'l', T: 298.15 K, P: 101325 Pa
     flow (kmol/hr): Water    16.5
                     Ethanol  0.139
                     Octanol  0.00409
    
    Assume that 1% of the feed is not in equilibrium (possibly due to poor mixing):
        
    >>> import thermosteam as tmo
    >>> tmo.settings.set_thermo(['Water', 'Ethanol', 'Octanol'], cache=True)
    >>> feed = tmo.Stream('feed', Water=20, Octanol=20, Ethanol=1)
    >>> top = tmo.Stream('top')
    >>> bottom = tmo.Stream('bottom')
    >>> ms = tmo.MultiStream('ms', phases='lL') # Store flow rate data here as well
    >>> tmo.separations.lle(feed, top, bottom, efficiency=0.99, multi_stream=ms)
    >>> ms.show()
    MultiStream: ms
     phases: ('L', 'l'), T: 298.15 K, P: 101325 Pa
     flow (kmol/hr): (L) Water    3.55
                         Ethanol  0.861
                         Octanol  20
                     (l) Water    16.5
                         Ethanol  0.139
                         Octanol  0.00408
    
    """
    if multi_stream:
        ms = multi_stream
        ms.copy_like(feed)
    else:
        ms = feed.copy()
    ms.lle(feed.T, top_chemical=top_chemical)
    top_phase = 'l'
    bottom_phase = 'L'
    if not top_chemical:
        rho_l = ms['l'].rho
        rho_L = ms['L'].rho
        top_L = rho_L < rho_l
        if top_L:
            top_phase = 'L'
            bottom_phase = 'l'
    top.mol[:] = ms.imol[top_phase]
    bottom.mol[:] = ms.imol[bottom_phase]
    top.T = bottom.T = feed.T
    top.P = bottom.P = feed.P
    if efficiency < 1.:
        top.mol *= efficiency
        bottom.mol *= efficiency
        mixing = (1. - efficiency) / 2. * feed.mol
        top.mol += mixing
        bottom.mol += mixing
        
def vle(feed, vap, liq, T=None, P=None, V=None, Q=None, x=None, y=None,
        multi_stream=None):
    """
    Run VLE mass and energy balance.
    
    Parameters
    ----------
    feed : Stream
        Mixed feed.
    vap : Stream
        Vapor fluid.
    liq : Stream
        Liquid fluid.
    P=None : float
        Operating pressure [Pa].
    Q=None : float
        Duty [kJ/hr].
    T=None : float
        Operating temperature [K].
    V=None : float
        Molar vapor fraction.
    x=None : float
        Molar composition of liquid (for binary mixtures).
    y=None : float
        Molar composition of vapor (for binary mixtures).
    multi_stream : MultiStream, optional
        Data from feed is passed to this stream to perform vapor-liquid equilibrium.
    
    Examples
    --------
    Perform vapor-liquid equilibrium on water and ethanol and split phases
    to vapor and liquid streams:
    
    >>> import thermosteam as tmo
    >>> tmo.settings.set_thermo(['Water', 'Ethanol'], cache=True)
    >>> feed = tmo.Stream('feed', Water=20, Ethanol=20)
    >>> vapor = tmo.Stream('top')
    >>> liquid = tmo.Stream('bottom')
    >>> tmo.separations.vle(feed, vapor, liquid, V=0.5, P=101325)
    >>> vapor.show()
    Stream: top
     phase: 'g', T: 353.88 K, P: 101325 Pa
     flow (kmol/hr): Water    7.72
                     Ethanol  12.3
    >>> liquid.show()
    Stream: bottom
     phase: 'l', T: 353.88 K, P: 101325 Pa
     flow (kmol/hr): Water    12.3
                     Ethanol  7.72
    
    It is also possible to save flow rate data in a multi-stream as well:
        
    >>> ms = tmo.MultiStream('ms', phases='lg')
    >>> tmo.separations.vle(feed, vapor, liquid, V=0.5, P=101325, multi_stream=ms)
    >>> ms.show()
    MultiStream: ms
     phases: ('g', 'l'), T: 353.88 K, P: 101325 Pa
     flow (kmol/hr): (g) Water    7.72
                         Ethanol  12.3
                     (l) Water    12.3
                         Ethanol  7.72
    
    """
    if multi_stream:
        ms = multi_stream
        ms.copy_like(feed)
    else:
        ms = feed.copy()
    H = feed.H + Q if Q is not None else None
    ms.vle(P=P, H=H, T=T, V=V, x=x, y=y)

    # Set Values
    vap.phase = 'g'
    liq.phase = 'l'
    vap.mol[:] = ms.imol['g']
    liq.mol[:] = ms.imol['l']
    vap.T = liq.T = ms.T
    vap.P = liq.P = ms.P
    
def material_balance(chemical_IDs, variable_inlets, constant_inlets=(),
                     constant_outlets=(), is_exact=True, balance='flow'):
    """
    Solve stream mass balance by iteration.
    
    Parameters
    ----------
    chemical_IDs : tuple[str]
        Chemicals that will be used to solve mass balance linear equations.
        The number of chemicals must be same as the number of input streams varied.
    variable_inlets : Iterable[Stream]
        Inlet streams that can vary in net flow rate to accomodate for the
        mass balance.
    constant_inlets: Iterable[Stream], optional
        Inlet streams that cannot vary in flow rates.
    constant_outlets: Iterable[Stream], optional
        Outlet streams that cannot vary in flow rates.
    is_exact=True : bool, optional
        True if exact flow rate solution is required for the specified IDs.
    balance='flow' : {'flow', 'composition'}, optional
          * 'flow': Satisfy output flow rates
          * 'composition': Satisfy net output molar composition
    
    Examples
    --------
    Vary inlet flow rates to satisfy outlet flow rates:
        
    >>> import thermosteam as tmo
    >>> tmo.settings.set_thermo(['Water', 'Ethanol'], cache=True)
    >>> in_a = tmo.Stream('in_a', Water=1)
    >>> in_b = tmo.Stream('in_b', Ethanol=1)
    >>> variable_inlets = [in_a, in_b]
    >>> in_c = tmo.Stream('in_c', Water=100)
    >>> constant_inlets = [in_c]
    >>> out_a = tmo.Stream('out_a', Water=200, Ethanol=2)
    >>> out_b = tmo.Stream('out_b', Ethanol=100)
    >>> constant_outlets = [out_a, out_b]
    >>> chemical_IDs = ('Water', 'Ethanol')
    >>> tmo.separations.material_balance(chemical_IDs, variable_inlets, constant_inlets, constant_outlets)
    >>> tmo.Stream.sum([in_a, in_b, in_c]).mol - tmo.Stream.sum([out_a, out_b]).mol # Molar flow rates entering and leaving are equal
    array([0., 0.])
    
    Vary inlet flow rates to satisfy outlet composition:
        
    >>> import thermosteam as tmo
    >>> tmo.settings.set_thermo(['Water', 'Ethanol'], cache=True)
    >>> in_a = tmo.Stream('in_a', Water=1)
    >>> in_b = tmo.Stream('in_b', Ethanol=1)
    >>> variable_inlets = [in_a, in_b]
    >>> in_c = tmo.Stream('in_c', Water=100)
    >>> constant_inlets = [in_c]
    >>> out_a = tmo.Stream('out_a', Water=200, Ethanol=2)
    >>> out_b = tmo.Stream('out_b', Ethanol=100)
    >>> constant_outlets = [out_a, out_b]
    >>> chemical_IDs = ('Water', 'Ethanol')
    >>> tmo.separations.material_balance(chemical_IDs, variable_inlets, constant_inlets, constant_outlets, balance='composition')
    >>> tmo.Stream.sum([in_a, in_b, in_c]).z_mol - tmo.Stream.sum([out_a, out_b]).z_mol # Molar composition entering and leaving are equal
    array([0., 0.])
    
    """
    # SOLVING BY ITERATION TAKES 15 LOOPS FOR 2 STREAMS
    # SOLVING BY LEAST-SQUARES TAKES 40 LOOPS
    solver = np.linalg.solve if is_exact else np.linalg.lstsq

    # Set up constant and variable streams
    if not variable_inlets:
        raise ValueError('variable_inlets must contain at least one stream')
    index = variable_inlets[0].chemicals.get_index(chemical_IDs)
    mol_out = sum([s.mol for s in constant_outlets])

    if balance == 'flow':
        # Perform the following calculation: Ax = b = f - g
        # Where:
        #    A = flow rate array
        #    x = factors
        #    b = target flow rates
        #    f = output flow rates
        #    g = constant inlet flow rates

        # Solve linear equations for mass balance
        A = np.array([s.mol for s in variable_inlets]).transpose()[index, :]
        f = mol_out[index]
        g = sum([s.mol[index] for s in constant_inlets])
        b = f - g
        x = solver(A, b)

        # Set flow rates for input streams
        for factor, s in zip(x, variable_inlets):
            s.mol[:] = s.mol * factor

    elif balance == 'composition':
        # Perform the following calculation:
        # Ax = b
        #    = sum( A_ * x_guess + g_ )f - g
        #    = A_ * x_guess * f - O
        # O  = sum(g_)*f - g
        # Where:
        # A_ is flow array for all species
        # g_ is constant flows for all species
        # Same variable definitions as in 'flow'

        # Set all variables
        A_ = np.array([s.mol for s in variable_inlets]).transpose()
        A = np.array([s.mol for s in variable_inlets]).transpose()[index, :]
        F_mol_out = mol_out.sum()
        z_mol_out = mol_out / F_mol_out if F_mol_out else mol_out
        f = z_mol_out[index]
        g_ = sum([s.mol for s in constant_inlets])
        g = g_[index]
        O = sum(g_) * f - g

        # Solve by iteration
        x_guess = np.ones_like(index)
        not_converged = True
        while not_converged:
            # Solve linear equations for mass balance
            b = (A_ * x_guess).sum()*f + O
            x_new = solver(A, b)
            infeasibles = x_new < 0.
            if infeasibles.any(): x_new -= x_new[infeasibles].min()
            denominator = x_guess.copy()
            denominator[denominator == 0.] = 1.
            not_converged = sum(((x_new - x_guess)/denominator)**2) > 1e-6
            x_guess = x_new

        # Set flow rates for input streams
        for factor, s in zip(x_new, variable_inlets):
            s.mol = s.mol * factor
    
    else:
        raise ValueError( "balance must be one of the following: 'flow', 'composition'")
        
# %% General functional algorithms based on MESH equations to solve multi-stage 
# liquid-liquid equilibrium.

@thermo_user
class StageLLE:
    __slots__ = ('feed', 'raffinate', 'solvent', 'extract', 'multi_stream',
                 'carrier_chemical', '_thermo', '_phi', '_K', '_IDs')
    
    def __init__(self, T=298.15, P=101325, feed=None,
                 solvent=None, thermo=None, carrier_chemical=None):
        self.feed = feed
        self.solvent = solvent
        thermo = self._load_thermo(thermo)
        self.multi_stream = multi_stream = tmo.MultiStream(
            None, T=T, P=P, phases=('l', 'L'), thermo=thermo
        )
        self.raffinate = multi_stream['l']
        self.extract = multi_stream['L']
        self.carrier_chemical = carrier_chemical
        
    @property
    def T(self):
        return self.multi_stream.T
    @T.setter
    def T(self, T):
        self.multi_stream.T = T
        
    @property
    def P(self):
        return self.multi_stream.P
    @P.setter
    def P(self, P):
        self.multi_stream.P = P
        
    def partition(self, partition_data=None):
        multi_stream = self.multi_stream
        multi_stream.mix_from([self.feed, self.solvent])
        lle = multi_stream.lle
        if partition_data:
            top = multi_stream['l']
            bottom = multi_stream['L']
            lle = multi_stream.lle
            phi = partition_data['phi']
            self._K = K = partition_data['K']
            self._IDs = IDs = partition_data['IDs']
            self._phi = partition(multi_stream, top, bottom, IDs, K, phi)
        else:
            lle(self.T, top_chemical=self.carrier_chemical or self.feed.main_chemical)
            self._IDs = tuple([i.ID for i in lle._lle_chemicals])
            self._phi = lle._phi
            self._K = lle._K 
        
    def balance_raffinate_flows(self):
        total_mol = self.feed.mol + self.solvent.mol
        extract_mol = self.extract.mol
        self.raffinate.mol[:] = total_mol - extract_mol
        
    @property
    def IDs(self):
        return self._IDs
    @property
    def phi(self):
        return self._phi
    @property
    def K(self):
        return self._K 
    
    def __repr__(self):
        return f"{type(self).__name__}(T={self.T}, P={self.P})"
    
@thermo_user
class MultiStageLLE:
    """
    Create a MultiStageLLE object that models a counter-current system
    of mixer-settlers for liquid-liquid extraction.
    
    Parameters
    ----------
    N_stages : int
        Number of stages.
    feed : Stream
        Feed with solute.
    solvent : Stream
        Solvent to contact feed and recover solute.
    carrier_chemical : str
        Name of main chemical in the feed (which is not selectively extracted by the solvent).
    partition_data : {'IDs': tuple[str], 'K': 1d array}, optional
        IDs of chemicals in equilibrium and partition coefficients (molar 
        composition ratio of the raffinate over the extract). If given,
        The mixer-settlers will be modeled with these constants. Otherwise,
        partition coefficients are computed based on temperature and composition.
    
    Examples
    --------
    Simulate 2-stage extraction of methanol from water using octanol:
    
    >>> import thermosteam as tmo
    >>> tmo.settings.set_thermo(['Water', 'Methanol', 'Octanol'], cache=True)
    >>> N_stages = 2
    >>> feed = tmo.Stream('feed', Water=500, Methanol=50)
    >>> solvent = tmo.Stream('solvent', Octanol=500)
    >>> stages = tmo.separations.MultiStageLLE(N_stages, feed, solvent)
    >>> stages.simulate_multi_stage_lle_without_side_draws()
    >>> stages.raffinate.show()
    Stream: 
     phase: 'l', T: 298.15 K, P: 101325 Pa
     flow (kmol/hr): Water     413
                     Methanol  8.4
                     Octanol   0.1
    >>> stages.extract.show()
    Stream: 
     phase: 'L', T: 298.15 K, P: 101325 Pa
     flow (kmol/hr): Water     87.
                     Methanol  41.
                     Octanol   500
    
    Simulate 10-stage extraction with user defined partition coefficients:
    
    >>> import numpy as np
    >>> tmo.settings.set_thermo(['Water', 'Methanol', 'Octanol'])
    >>> N_stages = 10
    >>> feed = tmo.Stream('feed', Water=5000, Methanol=500)
    >>> solvent = tmo.Stream('solvent', Octanol=5000)
    >>> stages = tmo.separations.MultiStageLLE(N_stages, feed, solvent,
    ...     partition_data={
    ...         'K': np.array([6.894, 0.7244, 3.381e-04]),
    ...         'IDs': ('Water', 'Methanol', 'Octanol'),
    ...         'phi': 0.4100271108219455 # Initial phase fraction guess. This is optional.
    ...     }
    ... )
    >>> stages.simulate_multi_stage_lle_without_side_draws()
    >>> stages.raffinate.show()
    Stream: 
     phase: 'l', T: 298.15 K, P: 101325 Pa
     flow (kmol/hr): Water     4.1e+03
                     Methanol  1.2
                     Octanol   1.5
    >>> stages.extract.show()
    Stream: 
     phase: 'L', T: 298.15 K, P: 101325 Pa
     flow (kmol/hr): Water     871
                     Methanol  499
                     Octanol   5e+03
    
    """
    __slots__ = ('stages', 'index', 'multi_stream', 
                 'carrier_chemical', 'extract_flow_rates', 
                 'partition_data', '_thermo')
    
    def __init__(self, N_stages, feed, solvent, carrier_chemical=None,
                 thermo=None, partition_data=None):
        thermo = self._load_thermo(thermo)
        self.multi_stream = tmo.MultiStream(None, phases=('l', 'L'), thermo=thermo)
        self.stages = stages = [StageLLE(thermo=thermo) for i in range(N_stages)]
        self.carrier_chemical = carrier_chemical
        self.partition_data = partition_data
        for i in range(N_stages-1):
            stage = stages[i]
            next_stage = stages[i + 1]
            next_stage.feed = stage.raffinate
            stage.solvent = next_stage.extract
        stages[0].feed = feed
        stages[-1].solvent = solvent
        
    def __len__(self):
        return len(self.stages)
    def __iter__(self):
        return iter(self.stages)
    def __getitem__(self, key):
        return self.stages[key]
    
    @property
    def feed(self):
        return self.stages[0].feed
    @property
    def extract(self):
        return self.stages[0].extract    
    @property
    def solvent(self):
        return self.stages[-1].solvent
    @property
    def raffinate(self):
        return self.stages[-1].raffinate
    
    def update_multi_stage_lle_without_side_draws(self, extract_flow_rates):
        if (extract_flow_rates < 0.).any(): raise RuntimeError('negative flow rate')
        index = self.index
        for stage, extract_flow in zip(self.stages, extract_flow_rates):
            stage.extract.mol[index] = extract_flow
            stage.balance_raffinate_flows()
    
    def simulate_multi_stage_lle_without_side_draws(self):
        f = self.multi_stage_lle_without_side_draws_iter
        if hasattr(self, 'extract_flow_rates'):
            extract_flow_rates = self.extract_flow_rates
        else:
            extract_flow_rates = self.initialize_multi_stage_lle_without_side_draws()
        extract_flow_rates = flx.wegstein(f, extract_flow_rates, xtol=0.1, maxiter=10, checkiter=False)
        self.extract_flow_rates = extract_flow_rates
        self.update_multi_stage_lle_without_side_draws(extract_flow_rates)
    
    def initialize_multi_stage_lle_without_side_draws(self):
        feed = self.feed
        solvent = self.solvent
        multi_stream = self.multi_stream
        multi_stream.mix_from([feed, solvent]) 
        stages = self.stages
        N_stages = len(stages)
        if self.partition_data:
            data = self.partition_data
            top = multi_stream['l']
            bottom = multi_stream['L']
            IDs = data['IDs']
            K = data['K']
            phi = data['phi']
            data['phi'] = phi = partition(multi_stream, top, bottom, IDs, K, phi)
        else:
            lle = multi_stream.lle
            lle(multi_stream.T, top_chemical=self.carrier_chemical or feed.main_chemical)
            IDs = tuple([i.ID for i in lle._lle_chemicals])
            K = lle._K
            phi = lle._phi
        index = multi_stream.chemicals.get_index(IDs)
        phase_fractions = np.ones(N_stages) * phi
        partition_coefficients = np.ones([K.size, N_stages]) * K[:, np.newaxis]
        extract_flow_rates = flow_rates_for_multi_stage_extration_without_side_draws(
            N_stages, phase_fractions, partition_coefficients, feed.mol[index], solvent.mol[index]
        )
        self.index = index 
        return extract_flow_rates
    
    def multi_stage_lle_without_side_draws_iter(self, extract_flow_rates):
        self.update_multi_stage_lle_without_side_draws(extract_flow_rates)
        stages = self.stages
        for i in stages: i.partition(self.partition_data)
        K = np.transpose([i.K for i in stages]) 
        phi = np.array([i.phi for i in stages])
        index = self.index
        stages = self.stages
        N_stages = len(stages)
        extract_flow_rates = flow_rates_for_multi_stage_extration_without_side_draws(
            N_stages, phi, K, self.feed.mol[index], self.solvent.mol[index]
        )
        return extract_flow_rates
        
@flx.njitable(cache=True)
def single_component_flow_rates_for_multi_stage_lle_without_side_draws(
        N_stages,
        phase_ratios,
        partition_coefficients, 
        feed, 
        solvent
    ):
    """
    Solve flow rates for a single component across a multi stage liquid-liquid
    extraction operation without side draws. 

    Parameters
    ----------
    N_stages : int
        Number of stages.
    phase_ratios : 1d array
        Phase ratios by stage. The phase ratio for a given stage is 
        defined as F_l / F_L; where F_l and F_L are the flow rates 
        of phase l (raffinate) and L (extract) leaving the stage respectively.
    partition_coefficients : 1d array
        Partition coefficients by stage. The partition coefficient for a given
        stage is defined as x_l / x_L; where x_l and x_L are the fraction of
        the component in phase l (raffinate) and L (extract) leaving the stage.
    feed : float
        Component flow rate in feed entering stage 1.
    solvent : float
        Component flow rate in solvent entering stage N.

    Returns
    -------
    extract_flow_rates : 1d array
        Extract component flow rates by stage.

    """
    component_ratios = phase_ratios * partition_coefficients
    A = np.eye(N_stages) * (1 + component_ratios) 
    for i in range(N_stages-1):
        i_next = i + 1
        A[i, i_next] = -1
        A[i_next, i] = -component_ratios[i]
    b = np.zeros(N_stages)
    b[0] = feed
    b[-1] += solvent
    return np.linalg.solve(A, b)

@flx.njitable(cache=True)
def flow_rates_for_multi_stage_extration_without_side_draws(
        N_stages,
        phase_fractions,
        partition_coefficients, 
        feed, 
        solvent
    ):
    """
    Solve flow rates for a single component across a multi stage liquid-liquid
    extraction without side draws. 

    Parameters
    ----------
    N_stages : int
        Number of stages.
    phase_fractions : 1d array
        Phase fractions by stage. The phase fraction for a given stage is 
        defined as F_l / (F_l + F_L); where F_l and F_L are the flow rates 
        of phase l (raffinate) and L (extract) leaving the stage respectively.
    partition_coefficients : Iterable[1d array]
        Partition coefficients with components by row and stages by column.
        The partition coefficient for a component in a given stage is defined 
        as x_l / x_L; where x_l and x_L are the fraction of the component in 
        phase l (raffinate) and L (extract) leaving the stage.
    feed : Iterable[float]
        Flow rates of all components in feed entering stage 1.
    solvent : Iterable[float]
        Flow rates of all components in solvent entering stage N.

    Returns
    -------
    extract_flow_rates : 2d array
        Extract flow rates with stages by row and components by column.

    """
    phase_ratios = phase_fractions / (1. - phase_fractions)
    N_chemicals = feed.size
    extract_flow_rates = np.zeros((N_stages, N_chemicals))
    for i in range(N_chemicals):
        flows_by_stage = single_component_flow_rates_for_multi_stage_lle_without_side_draws(
                N_stages, phase_ratios, partition_coefficients[i], feed[i], solvent[i]
        ) 
        extract_flow_rates[:, i] = flows_by_stage
    return extract_flow_rates



# -*- coding: utf-8 -*-
# BioSTEAM: The Biorefinery Simulation and Techno-Economic Analysis Modules
# Copyright (C) 2020, Yoel Cortes-Pena <yoelcortes@gmail.com>
# 
# This module extends the heat_capacity module from the chemicals's library:
# https://github.com/CalebBell/chemicals
# Copyright (C) 2020 Caleb Bell <Caleb.Andrew.Bell@gmail.com>
#
# This module is under a dual license:
# 1. The UIUC open-source license. See 
# github.com/BioSTEAMDevelopmentGroup/biosteam/blob/master/LICENSE.txt
# for license details.
# 
# 2. The MIT open-source license. See
# https://github.com/CalebBell/chemicals/blob/master/LICENSE.txt for details.
from chemicals import heat_capacity as hc
from math import log
import numpy as np
from ..utils import forward
from ..base import (InterpolatedTDependentModel,
                    TDependentHandleBuilder, 
                    PhaseTHandleBuilder, 
                    thermo_model, 
                    functor)
from .data import (
    VDI_saturation_dict, 
    lookup_VDI_tabular_data,
    Cp_data_Poling,
    TRC_gas_data,
    CRC_standard_data,
    Cp_dict_PerryI,
    zabransky_dict_sat_s,
    zabransky_dict_sat_p,
    zabransky_dict_const_s,
    zabransky_dict_const_p,
    zabransky_dict_iso_s,
    zabransky_dict_iso_p,
)
from ..constants import calorie

hc.__all__.extend([
    'heat_capacity_handle',
    
    'Lastovka_Shaw_definite_integral',
    'Lastovka_Shaw_definite_integral_over_T',

    'TRCCp_definite_integral', 
    'TRCCp_definite_integral_over_T',

    'Poling_definite_integral', 
    'Poling_definite_integral_over_T',

    'Zabransky_cubic_definite_integral', 
    'Zabransky_cubic_definite_integral_over_T',

    'Zabransky_quasi_polynomial_definite_integral', 
    'Zabransky_quasi_polynomial_definite_integral_over_T',
 
    'Dadgostar_Shaw_definite_integral', 
    'Dadgostar_Shaw_definite_integral_over_T',

    'Lastovka_solid_definite_integral', 
    'Lastovka_solid_definite_integral_over_T',
    
    'Perry_151_definite_integral', 
    'Perry_151_definite_integral_over_T'
])


# %% Utilities

def CnHS(FCn, FH, FS, data):
    fCn = FCn.from_args(data)
    return {'evaluate': fCn,
            'integrate_by_T': FH.from_other(fCn),
            'integrate_by_T_over_T': FS.from_other(fCn)}

def CnHSModel(FCn, FH, FS, data=None, Tmin=None, Tmax=None, name=None):
    funcs = CnHS(FCn, FH, FS, data)
    return thermo_model(Tmin=Tmin, Tmax=Tmax, name=name, var='Cn', **funcs)

class ZabranskyModelBuilder:
    __slots__ = ('name', 'casdata', 'funcs', 'many')
    
    def __init__(self, name, casdata, funcs, many=False):
        self.name = name
        self.casdata = casdata
        self.funcs = funcs
        self.many = many
    
    def add_model(self, CAS, models):
        if CAS in self.casdata:
            if self.many:
                models.extend([self.build_model(i) for i in self.casdata[CAS]])
            else:
                models.append(self.build_model(self.casdata[CAS]))
    
    def build_model(self, data):
        try:
            coeffs = (data.Tc, *data.coeffs)
        except:
            coeffs = data.coeffs
        return CnHSModel(*self.funcs, coeffs, data.Tmin, data.Tmax, self.name)


# %% Heat Capacities Gas

Lastovka_Shaw = functor(hc.Lastovka_Shaw, 'Cn.g')

@forward(hc)
@functor(var='H.g')
def Lastovka_Shaw_definite_integral(Ta, Tb, MW, similarity_variable, cyclic_aliphatic=False):
    term = hc.Lastovka_Shaw_term_A(similarity_variable, cyclic_aliphatic)
    return (hc.Lastovka_Shaw_integral(Tb, similarity_variable, cyclic_aliphatic, MW, term)
            - hc.Lastovka_Shaw_integral(Ta, similarity_variable, cyclic_aliphatic, MW, term))

@forward(hc)    
@functor(var='S.g')
def Lastovka_Shaw_definite_integral_over_T(Ta, Tb, MW, similarity_variable,
                                  cyclic_aliphatic=False):
    term = hc.Lastovka_Shaw_term_A(similarity_variable, cyclic_aliphatic)
    return (hc.Lastovka_Shaw_integral_over_T(Tb, similarity_variable, cyclic_aliphatic, MW, term)
            - hc.Lastovka_Shaw_integral_over_T(Ta, similarity_variable, cyclic_aliphatic, MW, term))

TRCCp = functor(hc.TRCCp, 'Cn.g')

@forward(hc)
@functor(var='H.g')
def TRCCp_definite_integral(Ta, Tb, a0, a1, a2, a3, a4, a5, a6, a7):
    return (hc.TRCCp_integral(Tb, a0, a1, a2, a3, a4, a5, a6, a7)
            - hc.TRCCp_integral(Ta, a0, a1, a2, a3, a4, a5, a6, a7))

@forward(hc)
@functor(var='S.g')
def TRCCp_definite_integral_over_T(Ta, Tb, a0, a1, a2, a3, a4, a5, a6, a7):
    return (hc.TRCCp_integral_over_T(Tb, a0, a1, a2, a3, a4, a5, a6, a7)
            - hc.TRCCp_integral_over_T(Ta, a0, a1, a2, a3, a4, a5, a6, a7))

Poling = functor(hc.Poling, 'Cn.g')

@forward(hc)
@functor(var='H.g')
def Poling_definite_integral(Ta, Tb, a, b, c, d, e):
    return (hc.Poling_integral(Tb, a, b, c, d, e)
            - hc.Poling_integral(Ta, a, b, c, d, e))

@forward(hc)    
@functor(var='S.g')
def Poling_definite_integral_over_T(Ta, Tb, a, b, c, d, e):
    return (hc.Poling_integral_over_T(Tb, a, b, c, d, e)
            - hc.Poling_integral_over_T(Ta, a, b, c, d, e))

# Heat capacity gas methods

TRCCp_functors = (TRCCp.functor,
                  TRCCp_definite_integral.functor,
                  TRCCp_definite_integral_over_T.functor)
Poling_functors = (Poling.functor,
                   Poling_definite_integral.functor, 
                   Poling_definite_integral_over_T.functor)

@TDependentHandleBuilder('Cn.g')
def heat_capacity_gas_handle(handle, CAS, MW, similarity_variable, cyclic_aliphatic):
    add_model = handle.add_model
    if CAS in TRC_gas_data:
        Tmin, Tmax, a0, a1, a2, a3, a4, a5, a6, a7, _, _, _ = TRC_gas_data[CAS]
        funcs = CnHS(*TRCCp_functors, (a0, a1, a2, a3, a4, a5, a6, a7))
        add_model(Tmin=Tmin, Tmax=Tmax, name=hc.TRCIG, **funcs)
    if CAS in Cp_data_Poling:
        Tmin, Tmax, a, b, c, d, e, Cn_g, _ = Cp_data_Poling[CAS]
        if not np.isnan(a):
            funcs = CnHS(*Poling_functors, (a, b, c, d, e))
            add_model(Tmin=Tmin, Tmax=Tmax, **funcs, name=hc.POLING)
        if not np.isnan(Cn_g):
            add_model(Cn_g, Tmin, Tmax, name=hc.POLING_CONST)
    if CAS in CRC_standard_data:
        Cn_g = CRC_standard_data[CAS][-1]
        if not np.isnan(Cn_g):
            add_model(Cn_g, name=hc.CRCSTD)
    if MW and similarity_variable:
        data = (similarity_variable, cyclic_aliphatic, MW)
        add_model(Lastovka_Shaw.functor.from_args(data), name=hc.LASTOVKA_SHAW)
    if CAS in VDI_saturation_dict:
        # NOTE: VDI data is for the saturation curve, i.e. at increasing
        # pressure; it is normally substantially higher than the ideal gas
        # value
        Ts, Cn_gs = lookup_VDI_tabular_data(CAS, 'Cp (g)')
        add_model(InterpolatedTDependentModel(Ts, Cn_gs, Tmin=Ts[0], Tmax=Ts[-1], name=hc.VDI_TABULAR))
hc.heat_capacity_gas_handle = heat_capacity_gas_handle
    
### Heat capacities of liquids

Rowlinson_Poling = functor(hc.Rowlinson_Poling, 'Cn.l')

def Rowlinson_Poling_hook(self, T, kwargs):
    Cpgm = kwargs['Cpgm']
    if callable(Cpgm): 
        kwargs = kwargs.copy()
        kwargs['Cpgm'] = Cpgm(T)
    return self.function(T, **kwargs)
Rowlinson_Poling.functor.hook = Rowlinson_Poling_hook

# # This method is less accurate than Rowlinson Poling; so we don't include it
# @forward(hc)
# @functor(var='Cn.l')
# def Rowlinson_Bondi_2(T, Tc, omega, Cpgm):
#     return hc.Rowlinson_Bondi(T, Tc, omega, Cpgm(T) if callable(Cpgm) else Cpgm)

Dadgostar_Shaw = functor(hc.Dadgostar_Shaw, 'Cn.l')

@forward(hc)
@functor(var='H.l')
def Dadgostar_Shaw_definite_integral(Ta, Tb, similarity_variable, MW):
    terms = hc.Dadgostar_Shaw_terms(similarity_variable)
    return (hc.Dadgostar_Shaw_integral(Tb, similarity_variable, MW, terms)
               - hc.Dadgostar_Shaw_integral(Ta, similarity_variable, MW, terms))
hc.Dadgostar_Shaw_definite_integral = Dadgostar_Shaw_definite_integral

@forward(hc)
@functor(var='S.l')
def Dadgostar_Shaw_definite_integral_over_T(Ta, Tb, similarity_variable, MW):
    terms = hc.Dadgostar_Shaw_terms(similarity_variable)
    return (hc.Dadgostar_Shaw_integral_over_T(Tb, similarity_variable, MW, terms)
               - hc.Dadgostar_Shaw_integral_over_T(Ta, similarity_variable, MW, terms))
hc.Dadgostar_Shaw_definite_integral_over_T = Dadgostar_Shaw_definite_integral_over_T

Zabransky_quasi_polynomial = functor(hc.Zabransky_quasi_polynomial, 'Cn.l')
 
@forward(hc)
@functor(var='H.l')
def Zabransky_quasi_polynomial_definite_integral(Ta, Tb, Tc, a1, a2, a3, a4, a5, a6):
    return (hc.Zabransky_quasi_polynomial_integral(Tb, Tc, a1, a2, a3, a4, a5, a6)
            - hc.Zabransky_quasi_polynomial_integral(Ta, Tc, a1, a2, a3, a4, a5, a6))
hc.Zabransky_quasi_polynomial_definite_integral = Zabransky_quasi_polynomial_definite_integral

@forward(hc)
@functor(var='S.l')
def Zabransky_quasi_polynomial_definite_integral_over_T(Ta, Tb, Tc, a1, a2, a3, a4, a5, a6):
    return (hc.Zabransky_quasi_polynomial_integral_over_T(Tb, Tc, a1, a2, a3, a4, a5, a6)
            - hc.Zabransky_quasi_polynomial_integral_over_T(Ta, Tc, a1, a2, a3, a4, a5, a6))
hc.Zabransky_quasi_polynomial_definite_integral_over_T = Zabransky_quasi_polynomial_definite_integral_over_T

Zabransky_cubic = functor(hc.Zabransky_cubic, 'Cn.l')

@forward(hc)
@functor(var='H.l')
def Zabransky_cubic_definite_integral(Ta, Tb, a1, a2, a3, a4):
    return (hc.Zabransky_cubic_integral(Tb, a1, a2, a3, a4)
            - hc.Zabransky_cubic_integral(Ta, a1, a2, a3, a4))

@forward(hc)
@functor(var='S.l')
def Zabransky_cubic_definite_integral_over_T(Ta, Tb, a1, a2, a3, a4):
    return (hc.Zabransky_cubic_integral_over_T(Tb, a1, a2, a3, a4)
            - hc.Zabransky_cubic_integral_over_T(Ta, a1, a2, a3, a4))
    
# Heat capacity liquid methods:
Zabransky_cubic_functors = (Zabransky_cubic.functor,
                            Zabransky_cubic_definite_integral.functor,
                            Zabransky_cubic_definite_integral_over_T.functor)
Zabransky_quasi_polynomial_functors = (Zabransky_quasi_polynomial.functor,
                                       Zabransky_quasi_polynomial_definite_integral.functor, 
                                       Zabransky_quasi_polynomial_definite_integral_over_T.functor)
Dadgostar_Shaw_functors = (Dadgostar_Shaw.functor,
                           Dadgostar_Shaw_definite_integral.functor,
                           Dadgostar_Shaw_definite_integral_over_T.functor)

zabransky_model_data = ((hc.ZABRANSKY_SPLINE,
                         zabransky_dict_const_s,
                         Zabransky_cubic_functors),
                        (hc.ZABRANSKY_QUASIPOLYNOMIAL,
                         zabransky_dict_const_p,
                         Zabransky_quasi_polynomial_functors),
                        (hc.ZABRANSKY_SPLINE_C,
                         zabransky_dict_iso_s, 
                         Zabransky_cubic_functors),
                        (hc.ZABRANSKY_QUASIPOLYNOMIAL_C, 
                         zabransky_dict_iso_p,
                         Zabransky_quasi_polynomial_functors),
                        (hc.ZABRANSKY_SPLINE_SAT,
                         zabransky_dict_sat_s,
                         Zabransky_cubic_functors),
                        (hc.ZABRANSKY_QUASIPOLYNOMIAL_SAT, 
                         zabransky_dict_sat_p,
                         Zabransky_quasi_polynomial_functors))

zabransky_model_builders = [ZabranskyModelBuilder(*i) for i in zabransky_model_data]
zabransky_model_builders[0].many = True
zabransky_model_builders[2].many = True
zabransky_model_builders[4].many = True

@TDependentHandleBuilder('Cn.l')
def heat_capacity_liquid_handle(handle, CAS, Tb, Tc, omega, MW, similarity_variable, Cn_g):
    for i in zabransky_model_builders: i.add_model(CAS, handle.models)        
    add_model = handle.add_model
    if CAS in VDI_saturation_dict:
        # NOTE: VDI data is for the saturation curve, i.e. at increasing
        # pressure; it is normally substantially higher than the ideal gas
        # value
        Ts, Cn_ls = lookup_VDI_tabular_data(CAS, 'Cp (l)')
        add_model(InterpolatedTDependentModel(Ts, Cn_ls, Ts[0], Ts[-1], name=hc.VDI_TABULAR))
    if Tc and omega and Cn_g:
        args = (Tc, omega, Cn_g, 200, Tc)
        add_model(Rowlinson_Poling.functor.from_args(args),Tmin=0, Tmax=Tc, name=hc.ROWLINSON_POLING)
    # Other
    if similarity_variable and MW:
        add_model(
            CnHSModel(*Dadgostar_Shaw_functors,
                      data=(similarity_variable, MW),
                      name=hc.DADGOSTAR_SHAW,
                      Tmin=0, Tmax=1e5)
        )
    # Constant models
    if CAS in Cp_data_Poling:
        Tmin, Tmax, a, b, c, d, e, Cn_g, Cn_l = Cp_data_Poling[CAS]
        if not np.isnan(Cn_g):
            if Tmax > 1e5: Tmax = 1e5
            add_model(Cn_l, Tmin, Tmax, name=hc.POLING_CONST)
    if CAS in CRC_standard_data:
        Cn_l = CRC_standard_data[CAS][-5]
        if not np.isnan(Cn_l):
            add_model(Cn_l, 0, Tc, name=hc.CRCSTD)
hc.heat_capacity_liquid_handle = heat_capacity_liquid_handle

# %% Heat Capacity Solid

Lastovka_solid = functor(hc.Lastovka_solid, 'Cn.s')

@forward(hc)
@functor(var='H.s')
def Lastovka_solid_definite_integral(Ta, Tb, similarity_variable, MW):
    return (hc.Lastovka_solid_integral(Tb, similarity_variable, MW)
            - hc.Lastovka_solid_integral(Ta, similarity_variable, MW))

@forward(hc)
@functor(var='S.s')
def Lastovka_solid_definite_integral_over_T(Ta, Tb, similarity_variable, MW):
    return (hc.Lastovka_solid_integral_over_T(Tb, similarity_variable, MW)
            - hc.Lastovka_solid_integral_over_T(Ta, similarity_variable, MW))
    
Perry_151 = functor(hc.Perry_151, 'Cn.s')

@forward(hc)
@functor(var='H.s')
def Perry_151_definite_integral(Ta, Tb, a, b, c, d):
    H1 = (a*Ta + 0.5*b*Ta**2 - d/Ta + c*Ta**3/3.)
    H2 = (a*Tb + 0.5*b*Tb**2 - d/Tb + c*Tb**3/3.)
    return (H2 - H1)*calorie

@forward(hc)
@functor(var='S.s')
def Perry_151_definite_integral_over_T(Ta, Tb, a, b, c, d):
    S1 = a*log(Ta) + b*Ta - d/(2.*Ta**2) + 0.5*c*Ta**2
    S2 = a*log(Tb) + b*Tb - d/(2.*Tb**2) + 0.5*c*Tb**2
    return (S2 - S1)*calorie

Lastovka_solid_functors = (Lastovka_solid.functor,
                           Lastovka_solid_definite_integral.functor,
                           Lastovka_solid_definite_integral_over_T.functor)
Perry_151_functors = (Perry_151.functor,
                      Perry_151_definite_integral.functor, 
                      Perry_151_definite_integral_over_T.functor)

# Heat capacity solid methods

@TDependentHandleBuilder('Cn.s')
def heat_capacity_solid_handle(handle, CAS, similarity_variable, MW):
    Tmin = 0
    Tmax = 2000
    add_model = handle.add_model
    if CAS in Cp_dict_PerryI:
        vals = Cp_dict_PerryI[CAS]
        if 'c' in vals:
            c = vals['c']
            Tmin = c['Tmin']
            Tmax = c['Tmax']
            data = (c['Const'], c['Lin'], c['Quad'], c['Quadinv'])
            add_model(CnHSModel(*Perry_151_functors, data), Tmin, Tmax,
                      name=hc.PERRY151)
    if CAS in CRC_standard_data:
        Cnc = CRC_standard_data[CAS][3]
        if not np.isnan(Cnc):
            add_model(float(Cnc), 200, 350)
    if similarity_variable and MW:
        data = (similarity_variable, MW)
        add_model(CnHSModel(*Lastovka_solid_functors, data), Tmin, Tmax,
                  name=hc.LASTOVKA_S)
hc.heat_capacity_solid_handle = heat_capacity_solid_handle

hc.heat_capacity_handle = PhaseTHandleBuilder(
    'Cn', heat_capacity_solid_handle, heat_capacity_liquid_handle, heat_capacity_gas_handle
)
